#!/usr/bin/env python3
# -*- encoding: Utf-8 -*-
import re
from typing import List, Tuple, Dict, Set, Sequence, Union, Optional, Any
from os.path import dirname, realpath

from hermes_dec.decompilation.defs import (
    HermesDecompiler,
    DecompiledFunctionBody,
    TokenString,
    LeftParenthesisToken,
    RightParenthesisToken,
    LeftHandRegToken,
    AssignmentToken,
    DotAccessorToken,
    IndexStringToken,
    CatchBlockStart,
    RightHandRegToken,
    GetEnvironmentToken,
    StoreToEnvironment,
    BindToken,
    SwitchImm,
    LoadFromEnvironmentToken,
    ReturnDirective,
    ResumeGenerator,
    SaveGenerator,
    StartGenerator,
    NewInnerEnvironmentToken,
    NewEnvironmentToken,
    JumpCondition,
    JumpNotCondition,
    FunctionTableIndex,
    ForInLoopInit,
    ForInLoopNextIter,
    ThrowDirective,
    ReturnDirective,
    RawToken,
)
from hermes_dec.parsers.serialized_literal_parser import (
    unpack_slp_array,
    SLPArray,
    SLPValue,
    TagType,
)
from hermes_dec.parsers.hbc_bytecode_parser import parse_hbc_bytecode


# e.g. 3D, *, "unsigned long long"
invalid_js_property = re.compile('^[^_a-zA-Z]')


def pass2_transform_code(
    state: HermesDecompiler, function_body: DecompiledFunctionBody
):

    TS = TokenString

    RD = ReturnDirective
    BIND = BindToken

    RG = ResumeGenerator
    SG = SaveGenerator
    SG2 = StartGenerator

    LHRT = LeftHandRegToken
    AT = AssignmentToken
    RHRT = RightHandRegToken
    RT = RawToken
    IST = IndexStringToken

    GET = GetEnvironmentToken
    STE = StoreToEnvironment
    NET = NewEnvironmentToken
    NIET = NewInnerEnvironmentToken
    LFET = LoadFromEnvironmentToken

    TD = ThrowDirective
    RD = ReturnDirective

    SI = SwitchImm
    JC = JumpCondition
    JNC = JumpNotCondition

    FILI = ForInLoopInit
    FILNI = ForInLoopNextIter

    DAT = DotAccessorToken
    CBS = CatchBlockStart

    """
    BFT = BuiltinFunctionToken
    ACT = AsyncClosureToken
    CT = ClosureToken
    FT = FunctionToken
    """
    FTI = FunctionTableIndex

    LPT = LeftParenthesisToken
    RPT = RightParenthesisToken

    lines = function_body.statements = []

    function_header = function_body.function_object

    for instruction in parse_hbc_bytecode(function_header, state.hbc_reader):
        op1 = getattr(instruction, 'arg1', None)
        op2 = getattr(instruction, 'arg2', None)
        op3 = getattr(instruction, 'arg3', None)
        op4 = getattr(instruction, 'arg4', None)
        op5 = getattr(instruction, 'arg5', None)
        op6 = getattr(instruction, 'arg6', None)

        if instruction.inst.name in ('Add', 'AddN'):
            lines.append(
                TS(
                    [LHRT(op1), AT(), RHRT(op2), RT(' + '), RHRT(op3)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == 'Add32':
            lines.append(
                TS(
                    [
                        LHRT(op1),
                        AT(),
                        RT('__uasm.add32'),
                        LPT(),
                        RHRT(op2),
                        RT(', '),
                        RHRT(op3),
                        RPT(),
                    ],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == 'AddEmptyString':
            lines.append(
                TS(
                    [LHRT(op1), AT(), RT("'' + "), RHRT(op2)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == 'AsyncBreakCheck':
            lines.append(TS([], assembly=[instruction]))
        elif instruction.inst.name == 'BitAnd':
            lines.append(
                TS(
                    [LHRT(op1), AT(), RHRT(op2), RT(' & '), RHRT(op3)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == 'BitNot':
            lines.append(
                TS(
                    [LHRT(op1), AT(), RT('~'), RHRT(op2)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == 'BitOr':
            lines.append(
                TS(
                    [LHRT(op1), AT(), RHRT(op2), RT(' | '), RHRT(op3)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == 'BitXor':
            lines.append(
                TS(
                    [LHRT(op1), AT(), RHRT(op2), RT(' ^ '), RHRT(op3)],
                    assembly=[instruction],
                )
            )
        # Note: It should be checked later whether the implementation of
        # parsing the "Call*" opcodes present below matches the difference
        # between the function_header.frameSize and the latest frame
        # register index across all version of the Hermes bytecode VM,
        # whether thisArg is handled correctly in all implementation of
        # the variants of the call, we should also eventually set up
        # unit tests, etc.
        elif instruction.inst.name in ('Call', 'CallLong'):
            args = []
            for register in reversed(
                range(
                    function_header.frameSize - 7 - op3,
                    function_header.frameSize - 7,
                )
            ):
                args += [RHRT(register), RT(', ')]
            lines.append(
                TS(
                    [
                        LHRT(op1),
                        AT(),
                        RHRT(function_header.frameSize - 7),
                        RT('['),
                        RHRT(op2),
                        RT(']'),
                        LPT(),
                        *args[:-1],
                        RPT(),
                    ],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name in ('Call1', 'Call2', 'Call3', 'Call4'):
            args = []
            for arg_count in range(1, int(instruction.inst.name[-1])):
                args += [RHRT([op4, op5, op6][arg_count - 1]), RT(', ')]
            lines.append(
                TS(
                    [
                        LHRT(op1),
                        AT(),
                        RHRT(op2),
                        BIND(op3),
                        LPT(),
                        *args[:-1],
                        RPT(),
                    ],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name in ('CallDirect', 'CallDirectLongIndex'):
            args = []
            for register in reversed(
                range(
                    function_header.frameSize - 7 - op2,
                    function_header.frameSize - 7,
                )
            ):
                args += [RHRT(register), RT(', ')]
            state.calldirect_function_ids.add(op3)
            lines.append(
                TS(
                    [
                        LHRT(op1),
                        AT(),
                        DAT(),
                        FTI(op3, state),
                        BIND(function_header.frameSize - 7),
                        LPT(),
                        *args[:-1],
                        RPT(),
                    ],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name in ('CallBuiltin', 'CallBuiltinLong'):
            args = []
            for register in reversed(
                range(
                    function_header.frameSize - 7 - (op3 - 1),
                    function_header.frameSize - 7,
                )
            ):
                args += [RHRT(register), RT(', ')]
            lines.append(
                TS(
                    [
                        LHRT(op1),
                        AT(),
                        FTI(op2, state, is_builtin=True),
                        LPT(),
                        *args[:-1],
                        RPT(),
                    ],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == 'Catch':
            lines.append(TS([CBS(op1)], assembly=[instruction]))
        elif instruction.inst.name == 'CoerceThisNS':
            lines.append(
                TS([LHRT(op1), AT(), RHRT(op2)], assembly=[instruction])
            )
        elif instruction.inst.name == 'CompleteGenerator':
            lines.append(TS([], assembly=[instruction]))
        elif instruction.inst.name in ('Construct', 'ConstructLong'):
            args = []
            for register in reversed(
                range(
                    function_header.frameSize - 7 - op3,
                    function_header.frameSize - 7,
                )
            ):
                args += [RHRT(register), RT(', ')]
            lines.append(
                TS(
                    [
                        LHRT(op1),
                        AT(),
                        RT('new '),
                        RHRT(function_header.frameSize - 7),
                        RT('['),
                        RHRT(op2),
                        RT(']'),
                        LPT(),
                        *args[:-1],
                        RPT(),
                    ],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name in (
            'CreateAsyncClosure',
            'CreateAsyncClosureLongIndex',
        ):
            lines.append(
                TS(
                    [
                        LHRT(op1),
                        AT(),
                        FTI(op3, state, op2, is_async=True, is_closure=True),
                    ],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name in (
            'CreateClosure',
            'CreateClosureLongIndex',
        ):
            lines.append(
                TS(
                    [LHRT(op1), AT(), FTI(op3, state, op2, is_closure=True)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == 'CreateEnvironment':
            lines.append(TS([NET(op1)], assembly=[instruction]))
        elif instruction.inst.name == 'CreateInnerEnvironment':
            lines.append(TS([NIET(op1, op2, op3)], assembly=[instruction]))
        elif instruction.inst.name in (
            'CreateGenerator',
            'CreateGeneratorLongIndex',
        ):
            lines.append(
                TS(
                    [LHRT(op1), AT(), FTI(op3, state, op2, is_generator=True)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name in (
            'CreateGeneratorClosure',
            'CreateGeneratorClosureLongIndex',
        ):
            lines.append(
                TS(
                    [
                        LHRT(op1),
                        AT(),
                        FTI(
                            op3, state, op2, is_closure=True, is_generator=True
                        ),
                    ],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == 'CreateRegExp':
            pattern_string = state.hbc_reader.strings[op2]
            flags_string = state.hbc_reader.strings[op3]

            pattern_string = pattern_string.translate(
                {char: repr(char) for char in range(0x20)}
            )
            #   pattern_string = pattern_string.replace('/', '\\/')

            readable_regexp = '/%s/%s' % (pattern_string, flags_string)

            lines.append(
                TS(
                    [LHRT(op1), AT(), RT(readable_regexp)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == 'CreateThis':
            lines.append(
                TS(
                    [
                        LHRT(op1),
                        AT(),
                        RT('Object.create'),
                        LPT(),
                        RHRT(op2),
                        RT(', {constructor: {value: '),
                        RHRT(op3),
                        RT('}}'),
                        RPT(),
                    ],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == 'Debugger':
            lines.append(TS([RT('debugger')], assembly=[instruction]))
        elif instruction.inst.name == 'DebuggerBreakCheck':
            lines.append(TS([], assembly=[instruction]))
        elif instruction.inst.name == 'Dec':
            lines.append(
                TS(
                    [LHRT(op1), AT(), RHRT(op2), RT(' - 1')],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == 'DeclareGlobalVar':
            global_var = state.hbc_reader.strings[op1]

            lines.append(
                TS(
                    [RT(global_var), AT(), RT('undefined')],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name in ('DelById', 'DelByIdLong'):
            prop_string = state.hbc_reader.strings[op3]

            lines.append(
                TS(
                    [
                        LHRT(op1),
                        AT(),
                        RT('delete '),
                        RHRT(op2),
                        DAT(),
                        RT(prop_string),
                    ],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == 'DelByVal':
            lines.append(
                TS(
                    [
                        LHRT(op1),
                        AT(),
                        RT('delete '),
                        RHRT(op2),
                        RT('['),
                        RHRT(op3),
                        RT(']'),
                    ],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == 'DirectEval':
            lines.append(
                TS(
                    [LHRT(op1), AT(), RT('eval'), LPT(), RHRT(op2), RPT()],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name in ('Div', 'DivN'):
            lines.append(
                TS(
                    [LHRT(op1), AT(), RHRT(op2), RT(' / '), RHRT(op3)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == 'Divi32':
            lines.append(
                TS(
                    [
                        LHRT(op1),
                        AT(),
                        RT('__uasm.divi32'),
                        LPT(),
                        RHRT(op2),
                        RT(', '),
                        RHRT(op3),
                        RPT(),
                    ],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == 'Divu32':
            lines.append(
                TS(
                    [
                        LHRT(op1),
                        AT(),
                        RT('__uasm.divu32'),
                        LPT(),
                        RHRT(op2),
                        RT(', '),
                        RHRT(op3),
                        RPT(),
                    ],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == 'Eq':
            lines.append(
                TS(
                    [LHRT(op1), AT(), RHRT(op2), RT(' == '), RHRT(op3)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == 'GetArgumentsLength':
            lines.append(
                TS(
                    [LHRT(op1), AT(), RT('arguments.length')],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == 'GetArgumentsPropByVal':
            lines.append(
                TS(
                    [
                        LHRT(op1),
                        AT(),
                        RT('arguments'),
                        RT('['),
                        RHRT(op2),
                        RT(']'),
                    ],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == 'GetBuiltinClosure':
            lines.append(
                TS(
                    [
                        LHRT(op1),
                        AT(),
                        FTI(op2, state, is_builtin=True, is_closure=True),
                    ],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name in (
            'GetById',
            'GetByIdLong',
            'GetByIdShort',
            'TryGetById',
            'TryGetByIdLong',
        ):
            string = state.hbc_reader.strings[op4]

            if ' ' in string or invalid_js_property.match(string):
                lines.append(
                    TS(
                        [LHRT(op1), AT(), RHRT(op2), IST(string)],
                        assembly=[instruction],
                    )
                )
            else:
                lines.append(
                    TS(
                        [LHRT(op1), AT(), RHRT(op2), DAT(), RT(string)],
                        assembly=[instruction],
                    )
                )

        elif instruction.inst.name == 'GetByVal':
            lines.append(
                TS(
                    [LHRT(op1), AT(), RHRT(op2), RT('['), RHRT(op3), RT(']')],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == 'GetEnvironment':
            if state.hbc_reader.header.version < 97:
                lines.append(TS([GET(op1, op2)], assembly=[instruction]))
            else:
                lines.append(TS([GET(op1, op3)], assembly=[instruction]))
        elif instruction.inst.name == 'GetGlobalObject':
            lines.append(
                TS([LHRT(op1), AT(), RT('global')], assembly=[instruction])
            )
        elif instruction.inst.name == 'GetNewTarget':
            lines.append(
                TS([LHRT(op1), AT(), RT('new.target')], assembly=[instruction])
            )
        elif instruction.inst.name == 'GetNextPName':
            lines.append(
                TS([FILNI(op1, op2, op3, op4, op5)], assembly=[instruction])
            )
        elif instruction.inst.name == 'GetPNameList':
            lines.append(
                TS([FILI(op1, op2, op3, op4)], assembly=[instruction])
            )
        elif instruction.inst.name == 'Greater':
            lines.append(
                TS(
                    [LHRT(op1), AT(), RHRT(op2), RT(' > '), RHRT(op3)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == 'GreaterEq':
            lines.append(
                TS(
                    [LHRT(op1), AT(), RHRT(op2), RT(' >= '), RHRT(op3)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == 'Inc':
            lines.append(
                TS(
                    [LHRT(op1), AT(), RHRT(op2), RT(' + 1')],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == 'InstanceOf':
            lines.append(
                TS(
                    [
                        LHRT(op1),
                        AT(),
                        RHRT(op2),
                        RT(' instanceof '),
                        RHRT(op3),
                    ],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == 'IsIn':
            lines.append(
                TS(
                    [LHRT(op1), AT(), RHRT(op2), RT(' in '), RHRT(op3)],
                    assembly=[instruction],
                )
            )
        # Note: The implementations of the methods related to the iteration below
        # is not fully consistent with the corresponding React Native instruction
        # behavior, which has special handling for iterating over arrays, may
        # suppress propagating the exception over certain IteratorClose() calls...
        #
        # This may be improved in the future if any required.
        elif instruction.inst.name == 'IteratorBegin':
            lines.append(
                TS(
                    [LHRT(op1), AT(), RHRT(op2), RT('[Symbol.iterator]')],
                    assembly=[instruction],
                )
            )
            lines.append(
                TS(
                    [
                        LHRT(op2),
                        AT(),
                        RHRT(op1),
                        LPT(),
                        RPT(),
                        DAT(),
                        RT('next'),
                    ],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == 'IteratorNext':
            lines.append(
                TS(
                    [
                        LHRT(op1),
                        AT(),
                        RHRT(op3),
                        LPT(),
                        RPT(),
                        DAT(),
                        RT('value'),
                    ],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == 'IteratorClose':
            lines.append(
                TS(
                    [RHRT(op1), DAT(), RT('return'), LPT(), RPT()],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name in ('JEqual', 'JEqualLong'):
            if op1 > 0:
                lines.append(
                    TS(
                        [
                            JNC(instruction.original_pos + op1),
                            RHRT(op2),
                            RT(' != '),
                            RHRT(op3),
                        ],
                        assembly=[instruction],
                    )
                )
            else:
                lines.append(
                    TS(
                        [
                            JC(instruction.original_pos + op1),
                            RHRT(op2),
                            RT(' == '),
                            RHRT(op3),
                        ],
                        assembly=[instruction],
                    )
                )
        elif instruction.inst.name in (
            'JGreater',
            'JGreaterN',
            'JGreaterLong',
            'JGreaterNLong',
        ):
            if op1 > 0:
                lines.append(
                    TS(
                        [
                            JNC(instruction.original_pos + op1),
                            RT('!'),
                            LPT(),
                            RHRT(op2),
                            RT(' > '),
                            RHRT(op3),
                            RPT(),
                        ],
                        assembly=[instruction],
                    )
                )
            else:
                lines.append(
                    TS(
                        [
                            JC(instruction.original_pos + op1),
                            RHRT(op2),
                            RT(' > '),
                            RHRT(op3),
                        ],
                        assembly=[instruction],
                    )
                )
        elif instruction.inst.name in (
            'JNotGreater',
            'JNotGreaterN',
            'JNotGreaterLong',
            'JNotGreaterNLong',
        ):
            if op1 > 0:
                lines.append(
                    TS(
                        [
                            JNC(instruction.original_pos + op1),
                            RHRT(op2),
                            RT(' > '),
                            RHRT(op3),
                        ],
                        assembly=[instruction],
                    )
                )
            else:
                lines.append(
                    TS(
                        [
                            JC(instruction.original_pos + op1),
                            RT('!'),
                            LPT(),
                            RHRT(op2),
                            RT(' > '),
                            RHRT(op3),
                            RPT(),
                        ],
                        assembly=[instruction],
                    )
                )
        elif instruction.inst.name in (
            'JGreaterEqual',
            'JGreaterEqualN',
            'JGreaterEqualLong',
            'JGreaterEqualNLong',
        ):
            if op1 > 0:
                lines.append(
                    TS(
                        [
                            JNC(instruction.original_pos + op1),
                            RT('!'),
                            LPT(),
                            RHRT(op2),
                            RT(' >= '),
                            RHRT(op3),
                            RPT(),
                        ],
                        assembly=[instruction],
                    )
                )
            else:
                lines.append(
                    TS(
                        [
                            JC(instruction.original_pos + op1),
                            RHRT(op2),
                            RT(' >= '),
                            RHRT(op3),
                        ],
                        assembly=[instruction],
                    )
                )
        elif instruction.inst.name in (
            'JNotGreaterEqual',
            'JNotGreaterEqualN',
            'JNotGreaterEqualLong',
            'JNotGreaterEqualNLong',
        ):
            if op1 > 0:
                lines.append(
                    TS(
                        [
                            JNC(instruction.original_pos + op1),
                            RHRT(op2),
                            RT(' >= '),
                            RHRT(op3),
                        ],
                        assembly=[instruction],
                    )
                )
            else:
                lines.append(
                    TS(
                        [
                            JC(instruction.original_pos + op1),
                            RT('!'),
                            LPT(),
                            RHRT(op2),
                            RT(' >= '),
                            RHRT(op3),
                            RPT(),
                        ],
                        assembly=[instruction],
                    )
                )
        elif instruction.inst.name in (
            'JLess',
            'JLessN',
            'JLessLong',
            'JLessNLong',
        ):
            if op1 > 0:
                lines.append(
                    TS(
                        [
                            JNC(instruction.original_pos + op1),
                            RT('!'),
                            LPT(),
                            RHRT(op2),
                            RT(' < '),
                            RHRT(op3),
                            RPT(),
                        ],
                        assembly=[instruction],
                    )
                )
            else:
                lines.append(
                    TS(
                        [
                            JC(instruction.original_pos + op1),
                            RHRT(op2),
                            RT(' < '),
                            RHRT(op3),
                        ],
                        assembly=[instruction],
                    )
                )
        elif instruction.inst.name in (
            'JNotLess',
            'JNotLessN',
            'JNotLessLong',
            'JNotLessNLong',
        ):
            if op1 > 0:
                lines.append(
                    TS(
                        [
                            JNC(instruction.original_pos + op1),
                            RHRT(op2),
                            RT(' < '),
                            RHRT(op3),
                        ],
                        assembly=[instruction],
                    )
                )
            else:
                lines.append(
                    TS(
                        [
                            JC(instruction.original_pos + op1),
                            RT('!'),
                            LPT(),
                            RHRT(op2),
                            RT(' < '),
                            RHRT(op3),
                            RPT(),
                        ],
                        assembly=[instruction],
                    )
                )
        elif instruction.inst.name in (
            'JLessEqual',
            'JLessEqualN',
            'JLessEqualLong',
            'JLessEqualNLong',
        ):
            if op1 > 0:
                lines.append(
                    TS(
                        [
                            JNC(instruction.original_pos + op1),
                            RT('!'),
                            LPT(),
                            RHRT(op2),
                            RT(' <= '),
                            RHRT(op3),
                            RPT(),
                        ],
                        assembly=[instruction],
                    )
                )
            else:
                lines.append(
                    TS(
                        [
                            JC(instruction.original_pos + op1),
                            RHRT(op2),
                            RT(' <= '),
                            RHRT(op3),
                        ],
                        assembly=[instruction],
                    )
                )
        elif instruction.inst.name in (
            'JNotLessEqual',
            'JNotLessEqualN',
            'JNotLessEqualLong',
            'JNotLessEqualNLong',
        ):
            if op1 > 0:
                lines.append(
                    TS(
                        [
                            JNC(instruction.original_pos + op1),
                            RHRT(op2),
                            RT(' <= '),
                            RHRT(op3),
                        ],
                        assembly=[instruction],
                    )
                )
            else:
                lines.append(
                    TS(
                        [
                            JC(instruction.original_pos + op1),
                            RT('!'),
                            LPT(),
                            RHRT(op2),
                            RT(' <= '),
                            RHRT(op3),
                            RPT(),
                        ],
                        assembly=[instruction],
                    )
                )
        elif instruction.inst.name in ('JNotEqual', 'JNotEqualLong'):
            if op1 > 0:
                lines.append(
                    TS(
                        [
                            JNC(instruction.original_pos + op1),
                            RHRT(op2),
                            RT(' == '),
                            RHRT(op3),
                        ],
                        assembly=[instruction],
                    )
                )
            else:
                lines.append(
                    TS(
                        [
                            JC(instruction.original_pos + op1),
                            RHRT(op2),
                            RT(' != '),
                            RHRT(op3),
                        ],
                        assembly=[instruction],
                    )
                )
        elif instruction.inst.name in ('JStrictEqual', 'JStrictEqualLong'):
            if op1 > 0:
                lines.append(
                    TS(
                        [
                            JNC(instruction.original_pos + op1),
                            RHRT(op2),
                            RT(' !== '),
                            RHRT(op3),
                        ],
                        assembly=[instruction],
                    )
                )
            else:
                lines.append(
                    TS(
                        [
                            JC(instruction.original_pos + op1),
                            RHRT(op2),
                            RT(' === '),
                            RHRT(op3),
                        ],
                        assembly=[instruction],
                    )
                )
        elif instruction.inst.name in (
            'JStrictNotEqual',
            'JStrictNotEqualLong',
        ):
            if op1 > 0:
                lines.append(
                    TS(
                        [
                            JNC(instruction.original_pos + op1),
                            RHRT(op2),
                            RT(' === '),
                            RHRT(op3),
                        ],
                        assembly=[instruction],
                    )
                )
            else:
                lines.append(
                    TS(
                        [
                            JC(instruction.original_pos + op1),
                            RHRT(op2),
                            RT(' !== '),
                            RHRT(op3),
                        ],
                        assembly=[instruction],
                    )
                )
        elif instruction.inst.name in ('Jmp', 'JmpLong'):
            if op1 > 0:
                lines.append(
                    TS(
                        [JNC(instruction.original_pos + op1), RT('false')],
                        assembly=[instruction],
                    )
                )
            else:
                lines.append(
                    TS(
                        [JC(instruction.original_pos + op1), RT('true')],
                        assembly=[instruction],
                    )
                )
        elif instruction.inst.name in ('JmpFalse', 'JmpFalseLong'):
            if op1 > 0:
                lines.append(
                    TS(
                        [JNC(instruction.original_pos + op1), RHRT(op2)],
                        assembly=[instruction],
                    )
                )
            else:
                lines.append(
                    TS(
                        [
                            JC(instruction.original_pos + op1),
                            RT('!'),
                            RHRT(op2),
                        ],
                        assembly=[instruction],
                    )
                )
        elif instruction.inst.name in ('JmpTrue', 'JmpTrueLong'):
            if op1 > 0:
                lines.append(
                    TS(
                        [
                            JNC(instruction.original_pos + op1),
                            RT('!'),
                            RHRT(op2),
                        ],
                        assembly=[instruction],
                    )
                )
            else:
                lines.append(
                    TS(
                        [JC(instruction.original_pos + op1), RHRT(op2)],
                        assembly=[instruction],
                    )
                )
        elif instruction.inst.name in ('JmpUndefined', 'JmpUndefinedLong'):
            if op1 > 0:
                lines.append(
                    TS(
                        [
                            JNC(instruction.original_pos + op1),
                            RHRT(op2),
                            RT(' !== undefined'),
                        ],
                        assembly=[instruction],
                    )
                )
            else:
                lines.append(
                    TS(
                        [
                            JC(instruction.original_pos + op1),
                            RHRT(op2),
                            RT(' === undefined'),
                        ],
                        assembly=[instruction],
                    )
                )
        elif instruction.inst.name == 'LShift':
            lines.append(
                TS(
                    [LHRT(op1), AT(), RHRT(op2), RT(' << '), RHRT(op3)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == 'Less':
            lines.append(
                TS(
                    [LHRT(op1), AT(), RHRT(op2), RT(' < '), RHRT(op3)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == 'LessEq':
            lines.append(
                TS(
                    [LHRT(op1), AT(), RHRT(op2), RT(' <= '), RHRT(op3)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name in (
            'LoadConstBigInt',
            'LoadConstBigIntLongIndex',
        ):
            lines.append(
                TS(
                    [
                        LHRT(op1),
                        AT(),
                        RT(str(state.hbc_reader.bigint_values[op2]) + 'n'),
                    ],
                    assembly=[instruction],
                )
            )  # TODO : Decode BigInts correctly
        elif instruction.inst.name in (
            'LoadConstDouble',
            'LoadConstInt',
            'LoadConstUInt8',
        ):
            lines.append(
                TS([LHRT(op1), AT(), str(op2)], assembly=[instruction])
            )
        elif instruction.inst.name in ('LoadConstEmpty', 'LoadConstUndefined'):
            lines.append(
                TS([LHRT(op1), AT(), RT('undefined')], assembly=[instruction])
            )
        elif instruction.inst.name == 'LoadConstFalse':
            lines.append(
                TS([LHRT(op1), AT(), RT('false')], assembly=[instruction])
            )
        elif instruction.inst.name == 'LoadConstNull':
            lines.append(
                TS([LHRT(op1), AT(), RT('null')], assembly=[instruction])
            )
        elif instruction.inst.name in (
            'LoadConstString',
            'LoadConstStringLongIndex',
        ):
            lines.append(
                TS(
                    [LHRT(op1), AT(), RT(repr(state.hbc_reader.strings[op2]))],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == 'LoadConstTrue':
            lines.append(
                TS([LHRT(op1), AT(), RT('true')], assembly=[instruction])
            )
        elif instruction.inst.name == 'LoadConstZero':
            lines.append(
                TS([LHRT(op1), AT(), RT('0')], assembly=[instruction])
            )
        elif instruction.inst.name in (
            'LoadFromEnvironment',
            'LoadFromEnvironmentL',
        ):
            lines.append(
                TS([LHRT(op1), AT(), LFET(op2, op3)], assembly=[instruction])
            )
        elif instruction.inst.name in ('LoadParam', 'LoadParamLong'):
            if not op2:
                arg_name = 'this'
            elif op2 < function_body.function_object.paramCount:
                arg_name = 'a' + str(op2 - 1)
            else:
                arg_name = 'arguments[%d]' % (op2 - 1)
            lines.append(
                TS([LHRT(op1), AT(), RT(arg_name)], assembly=[instruction])
            )
        elif instruction.inst.name == 'LoadThisNS':
            lines.append(
                TS([LHRT(op1), AT(), RT('this')], assembly=[instruction])
            )
        elif instruction.inst.name in (
            'Loadi16',
            'Loadi32',
            'Loadi8',
            'Loadu16',
            'Loadu32',
            'Loadu8',
        ):
            lines.append(
                TS(
                    [
                        LHRT(op1),
                        AT(),
                        RT('__uasm.' + instruction.inst.name.lower()),
                        LPT(),
                        RHRT(op2),
                        RT(', '),
                        RHRT(op3),
                        RPT(),
                    ],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == 'Mod':
            lines.append(
                TS(
                    [LHRT(op1), AT(), RHRT(op2), RT(' % '), RHRT(op3)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name in ('Mov', 'MovLong'):
            lines.append(
                TS([LHRT(op1), AT(), RHRT(op2)], assembly=[instruction])
            )
        elif instruction.inst.name == 'Mul32':
            lines.append(
                TS(
                    [
                        LHRT(op1),
                        AT(),
                        RT('__uasm.mul32'),
                        LPT(),
                        RHRT(op2),
                        RT(', '),
                        RHRT(op3),
                        RPT(),
                    ],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name in ('Mul', 'MulN'):
            lines.append(
                TS(
                    [LHRT(op1), AT(), RHRT(op2), RT(' * '), RHRT(op3)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == 'Negate':
            lines.append(
                TS(
                    [LHRT(op1), AT(), RT('-'), RHRT(op2)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == 'Neq':
            lines.append(
                TS(
                    [LHRT(op1), AT(), RHRT(op2), RT(' != '), RHRT(op3)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == 'NewArray':
            lines.append(
                TS(
                    [
                        LHRT(op1),
                        AT(),
                        RT('new Array'),
                        LPT(),
                        RT(str(op2)),
                        RPT(),
                    ],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name in (
            'NewArrayWithBuffer',
            'NewArrayWithBufferLong',
        ):
            array_text = '[%s]' % ', '.join(
                unpack_slp_array(
                    (
                        state.hbc_reader.arrays
                        if state.hbc_reader.header.version < 97
                        else state.hbc_reader.literal_values
                    )[op4:],
                    op3,
                ).to_strings(state.hbc_reader.strings)
            )
            lines.append(
                TS([LHRT(op1), AT(), RT(array_text)], assembly=[instruction])
            )
        elif instruction.inst.name == 'NewObject':
            lines.append(
                TS([LHRT(op1), AT(), RT('{}')], assembly=[instruction])
            )
        elif instruction.inst.name in (
            'NewObjectWithBuffer',
            'NewObjectWithBufferLong',
        ):
            if state.hbc_reader.header.version < 97:
                object_text = '{%s}' % ', '.join(
                    '%s: %s' % (key, value)
                    for key, value in zip(
                        unpack_slp_array(
                            state.hbc_reader.object_keys[op4:], op3
                        ).to_strings(state.hbc_reader.strings),
                        unpack_slp_array(
                            (state.hbc_reader.object_values)[op5:],
                            op3,
                        ).to_strings(state.hbc_reader.strings),
                    )
                )
            else:
                shape_keys = state.hbc_reader.object_shape_keys[op2]
                object_text = '{%s}' % ', '.join(
                    '%s: %s' % (key, value)
                    for key, value in zip(
                        shape_keys,
                        unpack_slp_array(
                            (state.hbc_reader.literal_values)[op3:],
                            len(shape_keys),
                        ).to_strings(state.hbc_reader.strings),
                    )
                )
            lines.append(
                TS([LHRT(op1), AT(), RT(object_text)], assembly=[instruction])
            )
        elif instruction.inst.name == 'NewObjectWithParent':
            lines.append(
                TS(
                    [
                        LHRT(op1),
                        AT(),
                        RT('Object.create'),
                        LPT(),
                        RHRT(op2),
                        RPT(),
                    ],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == 'Not':
            lines.append(
                TS(
                    [LHRT(op1), AT(), RT('!'), RHRT(op2)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == 'ProfilePoint':
            lines.append(TS([], assembly=[instruction]))
        elif instruction.inst.name in (
            'PutById',
            'PutByIdLong',
            'TryPutById',
            'TryPutByIdLong',
        ):
            index = repr(state.hbc_reader.strings[op4])
            lines.append(
                TS(
                    [LHRT(op1), RT('[' + index + ']'), AT(), RHRT(op2)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name in (
            'PutNewOwnById',
            'PutNewOwnByIdLong',
            'PutNewOwnByIdShort',
        ):
            index = repr(state.hbc_reader.strings[op3])
            lines.append(
                TS(
                    [LHRT(op1), RT('[' + index + ']'), AT(), RHRT(op2)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name in (
            'PutNewOwnNEById',
            'PutNewOwnNEByIdLong',
        ):
            index = repr(state.hbc_reader.strings[op3])
            lines.append(
                TS(
                    [
                        RT('Object.defineProperty'),
                        LPT(),
                        LHRT(op1),
                        RT(', '),
                        RT(repr(index)),
                        RT(', {value: '),
                        RHRT(op2),
                        RT('}'),
                        RPT(),
                    ],
                    assembly=[instruction],
                )
            )
            # TODO: Are non-enumerable values set correctly?
            # When are these used if they are used?
        elif instruction.inst.name == 'PutByVal':
            lines.append(
                TS(
                    [LHRT(op1), RT('['), RHRT(op2), RT(']'), AT(), RHRT(op3)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == 'PutOwnByVal':
            if op4:  # Is the property enumerable?
                lines.append(
                    TS(
                        [
                            LHRT(op1),
                            RT('['),
                            RHRT(op3),
                            RT(']'),
                            AT(),
                            RHRT(op2),
                        ],
                        assembly=[instruction],
                    )
                )
            else:
                lines.append(
                    TS(
                        [
                            RT('Object.defineProperty'),
                            LPT(),
                            LHRT(op1),
                            RT(', '),
                            RHRT(op3),
                            RT(', {value: '),
                            RHRT(op2),
                            RT('}'),
                            RPT(),
                        ],
                        assembly=[instruction],
                    )
                )
        elif instruction.inst.name in ('PutOwnByIndex', 'PutOwnByIndexL'):
            lines.append(
                TS(
                    [LHRT(op1), RT('[%d]' % op3), AT(), RHRT(op2)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name in ('PutOwnGetterSetterByVal'):
            index = repr(state.hbc_reader.strings[op3])
            lines.append(
                TS(
                    [
                        RT('Object.defineProperty'),
                        LPT(),
                        LHRT(op1),
                        RT(', '),
                        RHRT(op2),
                        RT(', {get: '),
                        RHRT(op3),
                        RT(', set: '),
                        RHRT(op4),
                        RT(
                            ', enumerable: '
                            + ('true' if op5 else 'false')
                            + '}'
                        ),
                        RPT(),
                    ],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == 'RShift':
            lines.append(
                TS(
                    [LHRT(op1), AT(), RHRT(op2), RT(' >> '), RHRT(op3)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == 'ReifyArguments':
            lines.append(
                TS([LHRT(op1), AT(), RT('arguments')], assembly=[instruction])
            )
        elif instruction.inst.name == 'ResumeGenerator':
            lines.append(TS([RG(op1, op2)], assembly=[instruction]))
        elif instruction.inst.name == 'Ret':
            lines.append(TS([RD(), RHRT(op1)], assembly=[instruction]))
        elif instruction.inst.name in ('SaveGenerator', 'SaveGeneratorLong'):
            lines.append(
                TS(
                    [SG(instruction.original_pos + op1)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == 'SelectObject':
            lines.append(
                TS(
                    [
                        LHRT(op1),
                        AT(),
                        RHRT(op3),
                        RT(' instanceof Object ? '),
                        RHRT(op3),
                        RT(' : '),
                        RHRT(op2),
                    ],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == 'StartGenerator':
            lines.append(TS([SG2()], assembly=[instruction]))
        elif instruction.inst.name in ('Store16', 'Store32', 'Store8'):
            lines.append(
                TS(
                    [
                        RT('__uasm.' + instruction.inst.name.lower()),
                        LPT(),
                        RHRT(op1),
                        RT(', '),
                        RHRT(op2),
                        RT(', '),
                        RHRT(op3),
                        RPT(),
                    ],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name in (
            'StoreNPToEnvironment',
            'StoreNPToEnvironmentL',
            'StoreToEnvironment',
            'StoreToEnvironmentL',
        ):
            lines.append(TS([STE(op1, op2, op3)], assembly=[instruction]))
        elif instruction.inst.name == 'StrictEq':
            lines.append(
                TS(
                    [LHRT(op1), AT(), RHRT(op2), RT(' === '), RHRT(op3)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == 'StrictNeq':
            lines.append(
                TS(
                    [LHRT(op1), AT(), RHRT(op2), RT(' !== '), RHRT(op3)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name in ('Sub', 'SubN'):
            lines.append(
                TS(
                    [LHRT(op1), AT(), RHRT(op2), RT(' - '), RHRT(op3)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == 'Sub32':
            lines.append(
                TS(
                    [
                        LHRT(op1),
                        AT(),
                        RT('__uasm.sub32'),
                        LPT(),
                        RHRT(op2),
                        RT(', '),
                        RHRT(op3),
                        RPT(),
                    ],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == 'SwitchImm':
            lines.append(
                TS(
                    [
                        SI(
                            op1,
                            instruction.original_pos + op2,
                            instruction.original_pos + op3,
                            op4,
                            op5,
                        ),
                        RT(
                            ' // Switch table: %s'
                            % instruction.switch_jump_table
                        ),
                    ],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == 'Throw':
            lines.append(TS([TD(), RHRT(op1)], assembly=[instruction]))
        elif instruction.inst.name == 'ThrowIfEmpty':
            lines.append(
                TS(
                    [
                        RT('if'),
                        LPT(),
                        RT('!'),
                        RHRT(op2),
                        RPT(),
                        RT(' '),
                        TD(),
                        RT(
                            'ReferenceError("accessing an uninitialized variable")'
                        ),
                    ],
                    assembly=[instruction],
                )
            )
            lines.append(
                TS([RT('else '), LHRT(op1), AT(), RHRT(op2)], assembly=[])
            )
        elif instruction.inst.name == 'ThrowIfHasRestrictedGlobalProperty':
            global_var = state.hbc_reader.strings[op1]

            lines.append(
                TS(
                    [
                        RT('if'),
                        LPT(),
                        RT('typeof global'),
                        DAT(),
                        RT(global_var),
                        RT(" === 'undefined'"),
                        RPT(),
                        RT(' '),
                        TD(),
                        RT(
                            'SyntaxError("Name is a restricted global identifier")'
                        ),
                    ],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == 'ThrowIfUndefinedInst':
            lines.append(
                TS(
                    [
                        RT('if'),
                        LPT(),
                        RT('typeof '),
                        RHRT(op1),
                        RT(" === 'undefined'"),
                        RPT(),
                        RT(' '),
                        TD(),
                        RT(
                            'ReferenceError("accessing an uninitialized variable")'
                        ),
                    ],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == 'ToInt32':
            lines.append(
                TS(
                    [LHRT(op1), AT(), RHRT(op2), RT(' | 0')],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == 'ToNumber':
            lines.append(
                TS(
                    [LHRT(op1), AT(), RHRT(op2), RT(' - 0')],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == 'ToNumeric':
            lines.append(
                TS(
                    [
                        LHRT(op1),
                        AT(),
                        RT('parseFloat'),
                        LPT(),
                        RHRT(op2),
                        RPT(),
                    ],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == 'TypeOf':
            lines.append(
                TS(
                    [LHRT(op1), AT(), RT('typeof '), RHRT(op2)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == 'URshift':
            lines.append(
                TS(
                    [LHRT(op1), AT(), RHRT(op2), RT(' >>> '), RHRT(op3)],
                    assembly=[instruction],
                )
            )
        elif instruction.inst.name == 'Unreachable':
            raise ValueError('Unreachable position reached: %r' % instruction)

        else:
            lines.append(
                TS(
                    [RT('// Unsupported instruction: %r' % instruction)],
                    assembly=[instruction],
                )
            )
            """
            # Todo: We should likely put an "Error comment" in the decompiled
            # source later, instead of this
            raise NotImplementedError('[Unsupported instruction in function "%s":] %r' % (
                function_body.function_name,
                instruction
            ))
            """



#!/usr/bin/env python3
# -*- encoding: Utf-8 -*-
from re import search, match, findall, sub, finditer, MULTILINE, DOTALL
from typing import List, Dict, Set, Sequence, Union, Optional, Any
from os.path import dirname, realpath, join
from dataclasses import dataclass
from html import escape

UTILS_DIR = dirname(realpath(__file__))
MODULE_DIR = dirname(realpath(UTILS_DIR))
SRC_DIR = dirname(realpath(MODULE_DIR))
ROOT_DIR = dirname(realpath(SRC_DIR))
DOCS_DIR = realpath(join(ROOT_DIR, 'docs'))

GIT_TAGS = 'v0.0.1 v0.0.3 v0.1.0 v0.2.1 hbc70 hbc73 v0.7.0 v0.8.0 v0.9.0 v0.12.0 hbc90 hbc92 hbc95 hbc97 hbc98 hbc99'.split(
    ' '
)

OUTPUT_FILE_NAME = DOCS_DIR + '/opcodes_table.html'


class OperandInfo:
    # Note for later: the extra information should maybe
    # put into an HTML tooltip (through the "title"
    # attribute) over respective icons, when rendering
    # the table cells regarding the bytecode versions
    # for each opcode and their respective changes

    operand_c_type: str
    operand_readable_type: str
    operand_meaning: Optional[str]
    operand_size_bytes: int


# Each column row should be laid in in the form
# of "<raw_opcode> (<struct_size_bytes>)", be
# bold if the structure size has changed from
# the previous version of the colum, display
# more information at hover


class ColumnInfo:
    raw_opcode: int
    present: bool
    has_changed_from_previous: bool
    has_ret_target: bool
    struct_size_bytes: int
    operands: List[OperandInfo]

    # Scrape comment information, and render it in a
    # separate row/tooltip then?
    plain_text_desc: Optional[str]

    def to_html(self):

        if self.has_changed_from_previous:
            html = '<td class="changed" '
        else:
            html = '<td '

        html += 'title="%s">' % self.plain_text_desc.replace('<br>', '\n')

        html += str(self.raw_opcode)

        # html += 'op=%s\x0asz=%s' % (
        #     self.raw_opcode,
        #     self.struct_size_bytes
        # )

        html += '</td>'

        return html


class InstructionRow:
    instruction_name: str
    bytecode_version_to_columns: Dict[int, ColumnInfo]


C_TYPE_TO_RAW_SIZE: Dict[str, int] = {
    'int8_t': 1,
    'int32_t': 4,
    'uint8_t': 1,
    'uint16_t': 2,
    'uint32_t': 4,
    'float': 4,
    'double': 8,
}


def main():

    instruction_name_to_row: Dict[str, InstructionRow] = {}

    all_bytecode_versions: Set[int] = set()

    for git_tag in GIT_TAGS:
        INPUT_FILE_NAME = (
            UTILS_DIR
            + '/original_hermes_bytecode_c_src/BytecodeList-%s.def' % git_tag
        )

        INPUT_VERSION_FILE_NAME = (
            UTILS_DIR
            + '/original_hermes_bytecode_c_src/BytecodeVersion-%s.h' % git_tag
        )

        with open(INPUT_VERSION_FILE_NAME) as fd:
            version_file_contents = fd.read()
        bytecode_version: int = int(
            search(r'BYTECODE_VERSION = (\d+)', version_file_contents).group(1)
        )

        all_bytecode_versions.add(bytecode_version)

        opcode_count = 0

        accumulated_comment: str = ''
        accumulated_comment_was_used: bool = False

        readable_type_to_c_type: Dict[str, str] = {}

        with open(INPUT_FILE_NAME) as fd:
            input_source = fd.read()

        # Backport OPERAND_FUNCTION_ID declarations added with
        # version 0.12.0 (https://github.com/facebook/hermes/commit/c20d7d8)
        # in order to improve disassembly output readability if needed.

        if 'OPERAND_FUNCTION_ID(CreateClosure, 3)' not in input_source:
            input_source += """
    OPERAND_FUNCTION_ID(CallDirect, 3)
    OPERAND_FUNCTION_ID(CreateClosure, 3)
    OPERAND_FUNCTION_ID(CreateClosureLongIndex, 3)
    """

            if 'CreateGeneratorClosure' in input_source:
                input_source += """
    OPERAND_FUNCTION_ID(CreateGeneratorClosure, 3)
    OPERAND_FUNCTION_ID(CreateGeneratorClosureLongIndex, 3)
    """

            if 'CreateGenerator' in input_source:
                input_source += """
    OPERAND_FUNCTION_ID(CreateGenerator, 3)
    OPERAND_FUNCTION_ID(CreateGeneratorLongIndex, 3)
    """

            if 'CreateAsyncClosure' in input_source:
                input_source += """
    OPERAND_FUNCTION_ID(CreateAsyncClosure, 3)
    OPERAND_FUNCTION_ID(CreateAsyncClosureLongIndex, 3)
    """

        lines: List[str] = input_source.splitlines()

        for line in lines:
            if not line.strip():
                accumulated_comment_was_used = False
                accumulated_comment = ''

            comment_line = match(r'^///\s*(.+)', line)

            if comment_line:
                comment = comment_line.group(1)

                if accumulated_comment_was_used:
                    accumulated_comment_was_used = False
                    accumulated_comment = ''
                accumulated_comment += comment.strip() + '\n'

            line = match(r'^((?:DEFINE|OPERAND)[^(]+?)\((.+?)\)', line)

            if line:
                directive, args = line.groups()
                args = args.split(', ')

                # print('=>', directive, args)

                def define_opcode(
                    instruction_name: str, operand_types: List[str]
                ):

                    nonlocal instruction_name_to_row
                    nonlocal opcode_count
                    nonlocal accumulated_comment_was_used
                    nonlocal accumulated_comment

                    instruction_row = instruction_name_to_row.setdefault(
                        instruction_name, InstructionRow()
                    )
                    instruction_row.instruction_name = instruction_name

                    if not getattr(
                        instruction_row, 'bytecode_version_to_columns', None
                    ):
                        instruction_row.bytecode_version_to_columns = {}

                    column_info = ColumnInfo()
                    instruction_row.bytecode_version_to_columns[
                        bytecode_version
                    ] = column_info

                    instruction_operands: List[OperandInfo] = []

                    for operand_readable_type in operand_types:
                        operand_info = OperandInfo()
                        operand_info.operand_meaning = None
                        operand_info.operand_readable_type = (
                            operand_readable_type
                        )
                        operand_info.operand_c_type = readable_type_to_c_type[
                            operand_readable_type
                        ]
                        operand_info.operand_size_bytes = C_TYPE_TO_RAW_SIZE[
                            operand_info.operand_c_type
                        ]

                        instruction_operands.append(operand_info)

                    column_info.has_changed_from_previous = False
                    column_info.struct_size_bytes = sum(
                        operand.operand_size_bytes
                        for operand in instruction_operands
                    )
                    column_info.present = True
                    column_info.raw_opcode = opcode_count
                    column_info.operands = instruction_operands

                    column_info.plain_text_desc = escape(
                        accumulated_comment.strip()
                    ).replace('\n', '<br>')

                    accumulated_comment_was_used = True

                    opcode_count += 1

                if directive.startswith('DEFINE_OPERAND_TYPE'):
                    readable_type_to_c_type[args[0]] = args[1]

                elif directive.startswith('DEFINE_OPCODE'):
                    define_opcode(args[0], args[1:])

                elif directive == 'DEFINE_JUMP_LONG_VARIANT':
                    pass

                elif directive.startswith('DEFINE_JUMP'):
                    args += {
                        'DEFINE_JUMP_1': ['Addr8'],
                        'DEFINE_JUMP_2': ['Addr8', 'Reg8'],
                        'DEFINE_JUMP_3': ['Addr8', 'Reg8', 'Reg8'],
                    }[directive]

                    saved_comment = accumulated_comment
                    define_opcode(args[0], args[1:])

                    accumulated_comment = saved_comment
                    define_opcode(args[0] + 'Long', ['Addr32'] + args[2:])

                elif directive.startswith('DEFINE_RET_TARGET'):
                    instruction_row = instruction_name_to_row[args[0]]

                    column_info = instruction_row.bytecode_version_to_columns[
                        bytecode_version
                    ]
                    column_info.has_ret_target = True

                elif directive.startswith('OPERAND_'):
                    operand_meaning = {
                        'OPERAND_BIGINT_ID': 'bigint_id',
                        'OPERAND_FUNCTION_ID': 'function_id',
                        'OPERAND_STRING_ID': 'string_id',
                    }[directive]

                    instruction_row = instruction_name_to_row[args[0]]

                    column_info = instruction_row.bytecode_version_to_columns[
                        bytecode_version
                    ]
                    column_info.operands[
                        int(args[1]) - 1
                    ].operand_meaning = operand_meaning

    # Diff each "ColumnInfo" entry contained within an
    # "InstructionRow" object, in order to set a
    # boolean value for the "has_changed_from_previous"
    # attribute of it

    for instruction_name, row in instruction_name_to_row.items():
        previous_column: Optional[ColumnInfo] = None

        for bytecode_version, column in sorted(
            row.bytecode_version_to_columns.items()
        ):
            column.plain_text_desc = '%s (total size %d)%s' % (
                ', '.join(
                    operand.operand_readable_type
                    if not operand.operand_meaning
                    else '%s (%s)'
                    % (operand.operand_readable_type, operand.operand_meaning)
                    for operand in column.operands
                ),
                column.struct_size_bytes,
                '<br><br>' + column.plain_text_desc
                if column.plain_text_desc
                else '',
            )

            if previous_column and (
                [operand.__dict__ for operand in previous_column.operands]
                != [operand.__dict__ for operand in column.operands]
                or column.plain_text_desc != previous_column.plain_text_desc
            ):
                column.has_changed_from_previous = True

            previous_column = column

    out_source = """<!DOCTYPE html>
    <html>
        <head>
            <title>Opcodes table</title>
            <meta charset="utf-8">
            <style>
                table {
                    border-collapse: collapse;
                    margin: 14px;
                }

                th, td {
                    padding: 3px;
                    border: 1px solid #ccc;
                }

                .changed {
                    background: #AB3535;
                }
            </style>
        </head>
        <body>
            <h1>Opcodes of the Hermes virtual machine</h1>
            <p>With the opcode byte value indicated respective to each stable version of the React Native Hermes VM bytecode format, as well as the auto-generated documentation from the "BytecodeList.def" file from the Hermes VM source tree.</p>
            <p>This list of opcodes is used by the <a href="https://github.com/P1sec/hermes-dec" target="_blank">hermes-dec</a> open-source reverse engineering project.</p>
            <table>
                <thead>
                    <tr>
                        <th>Instruction</th>
                        <th colspan="%s">Bytecode version to opcode</th>
                        <th>Documentation</th>
                    </tr>
                    <tr>
                        <th></th>
                        %s
                        <th></th>
                    </tr>
                </thead>
                <tbody>
    """ % (
        len(all_bytecode_versions),
        ''.join('<th>%s</th>' % tag for tag in sorted(all_bytecode_versions)),
    )

    for instruction_name, row in sorted(instruction_name_to_row.items()):
        out_source += (
            """<tr>
            <td>%s</td>"""
            % instruction_name
        )

        for bytecode_version in sorted(all_bytecode_versions):
            if bytecode_version in row.bytecode_version_to_columns:
                column = row.bytecode_version_to_columns[bytecode_version]
                out_source += column.to_html()
            else:
                out_source += '<td></td>'

        out_source += (
            """
            <td>%s</td>
        </tr>"""
            % column.plain_text_desc
        )

    out_source += """    </body>
    </html>
    """

    with open(OUTPUT_FILE_NAME, 'w', encoding='utf-8') as file_handle:
        file_handle.write(out_source)

    print()
    print('[+]  Wrote File => %s' % OUTPUT_FILE_NAME)

    print()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- encoding: Utf-8 -*-
from re import search, match, findall, sub, finditer, MULTILINE, DOTALL
from typing import List, Dict, Set, Sequence, Union, Optional, Any
from os.path import dirname, realpath, join
from dataclasses import dataclass
from os import listdir

UTILS_DIR = dirname(realpath(__file__))
MODULE_DIR = dirname(realpath(UTILS_DIR))
PARSERS_DIR = realpath(join(MODULE_DIR, 'parsers'))


def main():

    git_tags = sorted(
        file_name.split('-')[1].split('.def')[0]
        for file_name in listdir(
            UTILS_DIR + '/original_function_builtins_c_src/'
        )
        if file_name.endswith('.def')
    )

    for git_tag in git_tags:
        INPUT_FILE_NAME = (
            UTILS_DIR
            + '/original_hermes_bytecode_c_src/BytecodeList-%s.def' % git_tag
        )

        INPUT_BUILTINS_FILE_NAMES = (
            UTILS_DIR
            + '/original_function_builtins_c_src/Builtins-%s.def' % git_tag
        )

        INPUT_VERSION_FILE_NAME = (
            UTILS_DIR
            + '/original_hermes_bytecode_c_src/BytecodeVersion-%s.h' % git_tag
        )

        with open(INPUT_VERSION_FILE_NAME) as fd:
            version_file_contents = fd.read()
        bytecode_version: int = int(
            search(r'BYTECODE_VERSION = (\d+)', version_file_contents).group(1)
        )

        builtin_function_names: List[str] = []
        with open(INPUT_BUILTINS_FILE_NAMES) as fd:
            for line in fd:
                if line.split('(')[0] in (
                    'BUILTIN_METHOD',
                    'PRIVATE_BUILTIN',
                    'JS_BUILTIN',
                ):
                    operands = line.split('(')[1].split(')')[0].split(', ')
                    builtin_function_names.append('.'.join(operands))

        OUTPUT_FILE_NAME = (
            PARSERS_DIR + '/hbc_opcodes/hbc%d.py' % bytecode_version
        )

        out_source = '''#!/usr/bin/env python3
"""
    Note: The contents of the current file have been automatically
    generated by the "utils/hermes_bytecode_structs_parser.py"
    script

    Please do not edit it manually. 👍
"""

from typing import List, Set, Dict, Union, Optional, Sequence, Any

# Imports from the current diretory
from .def_classes import *

_instructions: List[Instruction] = []


'''

        opcode_count = 0

        with open(INPUT_FILE_NAME) as fd:
            input_source = fd.read()

        # Backport OPERAND_FUNCTION_ID declarations added with
        # version 0.12.0 (https://github.com/facebook/hermes/commit/c20d7d8)
        # in order to improve disassembly output readability if needed.

        if 'OPERAND_FUNCTION_ID(CreateClosure, 3)' not in input_source:
            input_source += """
OPERAND_FUNCTION_ID(CallDirect, 3)
OPERAND_FUNCTION_ID(CreateClosure, 3)
OPERAND_FUNCTION_ID(CreateClosureLongIndex, 3)
"""

            if 'CreateGeneratorClosure' in input_source:
                input_source += """
OPERAND_FUNCTION_ID(CreateGeneratorClosure, 3)
OPERAND_FUNCTION_ID(CreateGeneratorClosureLongIndex, 3)
"""

            if 'CreateGenerator' in input_source:
                input_source += """
OPERAND_FUNCTION_ID(CreateGenerator, 3)
OPERAND_FUNCTION_ID(CreateGeneratorLongIndex, 3)
"""

            if 'CreateAsyncClosure' in input_source:
                input_source += """
OPERAND_FUNCTION_ID(CreateAsyncClosure, 3)
OPERAND_FUNCTION_ID(CreateAsyncClosureLongIndex, 3)
"""

        lines = input_source.splitlines()

        for line in lines:
            line = match(r'^((?:DEFINE|OPERAND)[^(]+?)\((.+?)\)', line)

            if line:
                directive, args = line.groups()
                args = args.split(', ')

                # print('=>', directive, args)

                if directive.startswith('DEFINE_OPERAND_TYPE'):
                    out_source = out_source[
                        :-1
                    ]  # No endline before that for readability
                    out_source += f"{args[0]} = OperandType('{args[0]}', '{args[1]}')\n\n"
                elif directive.startswith('DEFINE_OPCODE'):
                    out_source += f"{args[0]} = Instruction('{args[0]}', {opcode_count}, [{', '.join(args[1:])}], globals())\n\n"
                    opcode_count += 1
                elif directive == 'DEFINE_JUMP_LONG_VARIANT':
                    pass
                elif directive.startswith('DEFINE_JUMP'):
                    args += {
                        'DEFINE_JUMP_1': ['Addr8'],
                        'DEFINE_JUMP_2': ['Addr8', 'Reg8'],
                        'DEFINE_JUMP_3': ['Addr8', 'Reg8', 'Reg8'],
                    }[directive]
                    out_source += f"{args[0]} = Instruction('{args[0]}', {opcode_count}, [{', '.join(args[1:])}], globals())\n"
                    out_source += f"{args[0]}Long = Instruction('{args[0]}Long', {opcode_count + 1}, [{', '.join(['Addr32'] + args[2:])}], globals())\n\n"
                    opcode_count += 2
                elif directive.startswith('DEFINE_RET_TARGET'):
                    out_source = out_source.strip()
                    out_source += f'\n{args[0]}.has_ret_target = True\n\n'
                elif directive.startswith('OPERAND_'):
                    operand_meaning = {
                        'OPERAND_BIGINT_ID': 'OperandMeaning.bigint_id',
                        'OPERAND_FUNCTION_ID': 'OperandMeaning.function_id',
                        'OPERAND_STRING_ID': 'OperandMeaning.string_id',
                    }[directive]
                    # out_source = out_source.strip()
                    out_source += f'{args[0]}.operands[{(int(args[1]) - 1)}].operand_meaning = {operand_meaning}\n\n'

        out_source += """_opcode_to_instruction : Dict[int, Instruction] = {v.opcode: v for v in _instructions}
_name_to_instruction : Dict[str, Instruction] = {v.name: v for v in _instructions}

"""

        out_source += """_builtin_function_names : List[str] = [
    %s
]

""" % ',\n    '.join(repr(string) for string in builtin_function_names)

        with open(OUTPUT_FILE_NAME, 'w', encoding='utf-8') as fd:
            fd.write(out_source)

        print()
        print('[+]  Wrote File => %s' % OUTPUT_FILE_NAME)

    print()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- encoding: Utf-8 -*-
from re import search, match, findall, sub, finditer, MULTILINE, DOTALL
from typing import List, Dict, Set, Sequence, Union, Optional, Any
from dataclasses import dataclass

from clang.cindex import Index, CursorKind

"""
    This file contains transitory libclang-based code
    that is used for generating structures which have
    been used as a base for the "../parsers/regexp_bytecode_parser.py"
    file, based on the original source code contained within
    the "../original_regex_bytecode_c_src" directory, which
    has been downloaded by the "get_source_codes.sh" script
    contained within the same directory.
    
    It is a scratchpad.
"""


@dataclass
class OpcodeStructureField:
    field_name: str
    field_type: str
    real_type: Optional[str]
    bit_len: Optional[int]


@dataclass
class OpcodeStructure:
    struct_name: str
    fields: List[OpcodeStructureField]


opcode_name_to_structure: Dict[str, OpcodeStructure] = {}


def find_typerefs(
    node,
    indent=0,
    struct: Optional[OpcodeStructure] = None,
    field: Optional[OpcodeStructureField] = None,
    visiting_baseclass: bool = False,
):
    """Find all references to the type named 'typename'"""
    # help(node.type)
    # exit()
    # help(node.type.get_declaration())
    # exit()
    """
    print('%sFound `%s` (%s) [line=%s, col=%s] %s' % (
        ' ' * (indent * 2),
        node.spelling, node.kind, node.location.line, node.location.column, (node.type.spelling, node.type.get_canonical().spelling, node.type.get_declaration().underlying_typedef_type.spelling if node.type.get_declaration().kind.is_declaration() else None))) # (node.storage_class, node.type, node.type.spelling, node.translation_unit, node.referenced, node.raw_comment)))
    """

    if node.kind == CursorKind.CXX_BASE_SPECIFIER:
        visiting_baseclass = True
    elif node.kind == CursorKind.STRUCT_DECL and (
        (node.spelling.endswith('Insn') and node.spelling != 'Insn')
        or 'BracketRange' in node.spelling
    ):
        struct = OpcodeStructure(node.spelling, [])
        opcode_name_to_structure[node.spelling] = struct
    elif struct and node.kind == CursorKind.FIELD_DECL:
        # help(node)
        # exit()
        field = OpcodeStructureField(
            node.spelling, node.type.spelling, None, None
        )
        if node.is_bitfield():
            field.bit_len = node.get_bitfield_width()
        if node.type.get_declaration().kind.is_declaration():
            underlying_type = (
                node.type.get_declaration().underlying_typedef_type.spelling
            )
            if (
                underlying_type
                and underlying_type != node.spelling
                and '__' not in underlying_type
            ):
                field.field_type, field.real_type = (
                    underlying_type,
                    field.field_type,
                )
        struct.fields.append(field)
    elif (
        struct
        and visiting_baseclass
        and node.kind == CursorKind.TYPE_REF
        and node.spelling != 'struct hermes::regex::Insn'
    ):
        for otherchild in node.type.get_declaration().get_children():
            find_typerefs(otherchild, indent + 1, struct, field, False)
    # Recurse for children of this node
    for child in node.get_children():
        find_typerefs(child, indent + 1, struct, field, visiting_baseclass)


# INPUT_FILE_NAME = '/home/marin/hermes/include/hermes/Regex/RegexBytecode.h'
INPUT_FILE_NAME = '/home/marin/hermes-dec/src/hermes_dec/parsers/original_regex_bytecode_c_src/RegexBytecode-v0.12.0.h'
# INPUT_FILE_NAME = '/home/marin/hermes-dec/src/hermes_dec/parsers/original_regex_bytecode_c_src/RegexBytecode-v0.0.1.h'


def main():

    index = Index.create()
    tu = index.parse(INPUT_FILE_NAME, args=['-x', 'c++'])
    print('Translation unit:', tu.spelling)
    find_typerefs(tu.cursor)

    print()
    print('=' * 12)
    print()
    print('=== Decoded structures: ===')
    print()

    print()
    for opcode_name, structure in opcode_name_to_structure.items():
        print('class %s(LittleEndianStructure):' % opcode_name)
        print('    _pack_ = True')
        print("    _layout _ = 'ms'")
        if structure.fields:
            print('    _fields_ = [')
            for field_index, field in enumerate(structure.fields):
                readable_type = {
                    'bool': 'c_bool',
                    'char': 'c_char',
                    'char16_t': 'c_wchar',
                    'uint8_t': 'c_uint8',
                    'uint16_t': 'c_uint16',
                    'uint32_t': 'c_uint32',
                }[field.field_type]

                if field.bit_len:
                    readable_type += ', %d' % field.bit_len

                if field.real_type:
                    print(
                        "        ('%s', %s)%s # %s"
                        % (
                            field.field_name,
                            readable_type,
                            ','
                            if field_index + 1 < len(structure.fields)
                            else '',
                            field.real_type,
                        )
                    )
                else:
                    print(
                        "        ('%s', %s)%s"
                        % (
                            field.field_name,
                            readable_type,
                            ','
                            if field_index + 1 < len(structure.fields)
                            else '',
                        )
                    )
            print('    ]')
        print()

    parsed_inst_names = sorted(
        set(
            opcode_name.replace('Insn', '')
            for opcode_name in opcode_name_to_structure.keys()
        )
        - {'BracketRange32'}
    )

    print('=')
    print()

    print('~')


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- encoding: Utf-8 -*-
from os.path import dirname, realpath
from os import execlp, chdir
import sys

SCRIPT_DIR = dirname(realpath(__file__))
sys.path.append(SCRIPT_DIR)


def main():
    chdir(SCRIPT_DIR)

    execlp('bash', 'bash', './regen_hbc_opcodes.sh', *sys.argv[1:])


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
"""
Note: The contents of the current file have been automatically
generated by the "utils/hermes_bytecode_structs_parser.py"
script

Please do not edit it manually. 👍
"""

from typing import List, Set, Dict, Union, Optional, Sequence, Any

# Imports from the current diretory
from .def_classes import *

_instructions: List[Instruction] = []

Reg8 = OperandType('Reg8', 'uint8_t')
Reg32 = OperandType('Reg32', 'uint32_t')
UInt8 = OperandType('UInt8', 'uint8_t')
UInt16 = OperandType('UInt16', 'uint16_t')
UInt32 = OperandType('UInt32', 'uint32_t')
Addr8 = OperandType('Addr8', 'int8_t')
Addr32 = OperandType('Addr32', 'int32_t')
Imm32 = OperandType('Imm32', 'int32_t')
Double = OperandType('Double', 'double')

Unreachable = Instruction('Unreachable', 0, [], globals())

NewObjectWithBuffer = Instruction(
    'NewObjectWithBuffer', 1, [Reg8, UInt16, UInt16, UInt16, UInt16], globals()
)

NewObjectWithBufferLong = Instruction(
    'NewObjectWithBufferLong',
    2,
    [Reg8, UInt16, UInt16, UInt32, UInt32],
    globals(),
)

NewObject = Instruction('NewObject', 3, [Reg8], globals())

NewObjectWithParent = Instruction(
    'NewObjectWithParent', 4, [Reg8, Reg8], globals()
)

NewArrayWithBuffer = Instruction(
    'NewArrayWithBuffer', 5, [Reg8, UInt16, UInt16, UInt16], globals()
)

NewArrayWithBufferLong = Instruction(
    'NewArrayWithBufferLong', 6, [Reg8, UInt16, UInt16, UInt32], globals()
)

NewArray = Instruction('NewArray', 7, [Reg8, UInt16], globals())

Mov = Instruction('Mov', 8, [Reg8, Reg8], globals())

MovLong = Instruction('MovLong', 9, [Reg32, Reg32], globals())

Negate = Instruction('Negate', 10, [Reg8, Reg8], globals())

Not = Instruction('Not', 11, [Reg8, Reg8], globals())

BitNot = Instruction('BitNot', 12, [Reg8, Reg8], globals())

TypeOf = Instruction('TypeOf', 13, [Reg8, Reg8], globals())

Eq = Instruction('Eq', 14, [Reg8, Reg8, Reg8], globals())

StrictEq = Instruction('StrictEq', 15, [Reg8, Reg8, Reg8], globals())

Neq = Instruction('Neq', 16, [Reg8, Reg8, Reg8], globals())

StrictNeq = Instruction('StrictNeq', 17, [Reg8, Reg8, Reg8], globals())

Less = Instruction('Less', 18, [Reg8, Reg8, Reg8], globals())

LessEq = Instruction('LessEq', 19, [Reg8, Reg8, Reg8], globals())

Greater = Instruction('Greater', 20, [Reg8, Reg8, Reg8], globals())

GreaterEq = Instruction('GreaterEq', 21, [Reg8, Reg8, Reg8], globals())

Add = Instruction('Add', 22, [Reg8, Reg8, Reg8], globals())

AddN = Instruction('AddN', 23, [Reg8, Reg8, Reg8], globals())

Mul = Instruction('Mul', 24, [Reg8, Reg8, Reg8], globals())

MulN = Instruction('MulN', 25, [Reg8, Reg8, Reg8], globals())

Div = Instruction('Div', 26, [Reg8, Reg8, Reg8], globals())

DivN = Instruction('DivN', 27, [Reg8, Reg8, Reg8], globals())

Mod = Instruction('Mod', 28, [Reg8, Reg8, Reg8], globals())

Sub = Instruction('Sub', 29, [Reg8, Reg8, Reg8], globals())

SubN = Instruction('SubN', 30, [Reg8, Reg8, Reg8], globals())

LShift = Instruction('LShift', 31, [Reg8, Reg8, Reg8], globals())

RShift = Instruction('RShift', 32, [Reg8, Reg8, Reg8], globals())

URshift = Instruction('URshift', 33, [Reg8, Reg8, Reg8], globals())

BitAnd = Instruction('BitAnd', 34, [Reg8, Reg8, Reg8], globals())

BitXor = Instruction('BitXor', 35, [Reg8, Reg8, Reg8], globals())

BitOr = Instruction('BitOr', 36, [Reg8, Reg8, Reg8], globals())

InstanceOf = Instruction('InstanceOf', 37, [Reg8, Reg8, Reg8], globals())

IsIn = Instruction('IsIn', 38, [Reg8, Reg8, Reg8], globals())

GetEnvironment = Instruction('GetEnvironment', 39, [Reg8, UInt8], globals())

StoreToEnvironment = Instruction(
    'StoreToEnvironment', 40, [Reg8, UInt8, Reg8], globals()
)

StoreToEnvironmentL = Instruction(
    'StoreToEnvironmentL', 41, [Reg8, UInt16, Reg8], globals()
)

StoreNPToEnvironment = Instruction(
    'StoreNPToEnvironment', 42, [Reg8, UInt8, Reg8], globals()
)

StoreNPToEnvironmentL = Instruction(
    'StoreNPToEnvironmentL', 43, [Reg8, UInt16, Reg8], globals()
)

LoadFromEnvironment = Instruction(
    'LoadFromEnvironment', 44, [Reg8, Reg8, UInt8], globals()
)

LoadFromEnvironmentL = Instruction(
    'LoadFromEnvironmentL', 45, [Reg8, Reg8, UInt16], globals()
)

GetGlobalObject = Instruction('GetGlobalObject', 46, [Reg8], globals())

GetNewTarget = Instruction('GetNewTarget', 47, [Reg8], globals())

CreateEnvironment = Instruction('CreateEnvironment', 48, [Reg8], globals())

DeclareGlobalVar = Instruction('DeclareGlobalVar', 49, [UInt32], globals())

DeclareGlobalVar.operands[0].operand_meaning = OperandMeaning.string_id

GetByIdShort = Instruction(
    'GetByIdShort', 50, [Reg8, Reg8, UInt8, UInt8], globals()
)

GetById = Instruction('GetById', 51, [Reg8, Reg8, UInt8, UInt16], globals())

GetByIdLong = Instruction(
    'GetByIdLong', 52, [Reg8, Reg8, UInt8, UInt32], globals()
)

GetByIdShort.operands[3].operand_meaning = OperandMeaning.string_id

GetById.operands[3].operand_meaning = OperandMeaning.string_id

GetByIdLong.operands[3].operand_meaning = OperandMeaning.string_id

TryGetById = Instruction(
    'TryGetById', 53, [Reg8, Reg8, UInt8, UInt16], globals()
)

TryGetByIdLong = Instruction(
    'TryGetByIdLong', 54, [Reg8, Reg8, UInt8, UInt32], globals()
)

TryGetById.operands[3].operand_meaning = OperandMeaning.string_id

TryGetByIdLong.operands[3].operand_meaning = OperandMeaning.string_id

PutById = Instruction('PutById', 55, [Reg8, Reg8, UInt8, UInt16], globals())

PutByIdLong = Instruction(
    'PutByIdLong', 56, [Reg8, Reg8, UInt8, UInt32], globals()
)

PutById.operands[3].operand_meaning = OperandMeaning.string_id

PutByIdLong.operands[3].operand_meaning = OperandMeaning.string_id

TryPutById = Instruction(
    'TryPutById', 57, [Reg8, Reg8, UInt8, UInt16], globals()
)

TryPutByIdLong = Instruction(
    'TryPutByIdLong', 58, [Reg8, Reg8, UInt8, UInt32], globals()
)

TryPutById.operands[3].operand_meaning = OperandMeaning.string_id

TryPutByIdLong.operands[3].operand_meaning = OperandMeaning.string_id

PutNewOwnByIdShort = Instruction(
    'PutNewOwnByIdShort', 59, [Reg8, Reg8, UInt8], globals()
)

PutNewOwnById = Instruction(
    'PutNewOwnById', 60, [Reg8, Reg8, UInt16], globals()
)

PutNewOwnByIdLong = Instruction(
    'PutNewOwnByIdLong', 61, [Reg8, Reg8, UInt32], globals()
)

PutNewOwnByIdShort.operands[2].operand_meaning = OperandMeaning.string_id

PutNewOwnById.operands[2].operand_meaning = OperandMeaning.string_id

PutNewOwnByIdLong.operands[2].operand_meaning = OperandMeaning.string_id

PutNewOwnNEById = Instruction(
    'PutNewOwnNEById', 62, [Reg8, Reg8, UInt16], globals()
)

PutNewOwnNEByIdLong = Instruction(
    'PutNewOwnNEByIdLong', 63, [Reg8, Reg8, UInt32], globals()
)

PutNewOwnNEById.operands[2].operand_meaning = OperandMeaning.string_id

PutNewOwnNEByIdLong.operands[2].operand_meaning = OperandMeaning.string_id

PutOwnByIndex = Instruction(
    'PutOwnByIndex', 64, [Reg8, Reg8, UInt8], globals()
)

PutOwnByIndexL = Instruction(
    'PutOwnByIndexL', 65, [Reg8, Reg8, UInt32], globals()
)

PutOwnByVal = Instruction(
    'PutOwnByVal', 66, [Reg8, Reg8, Reg8, UInt8], globals()
)

DelById = Instruction('DelById', 67, [Reg8, Reg8, UInt16], globals())

DelByIdLong = Instruction('DelByIdLong', 68, [Reg8, Reg8, UInt32], globals())

DelById.operands[2].operand_meaning = OperandMeaning.string_id

DelByIdLong.operands[2].operand_meaning = OperandMeaning.string_id

GetByVal = Instruction('GetByVal', 69, [Reg8, Reg8, Reg8], globals())

PutByVal = Instruction('PutByVal', 70, [Reg8, Reg8, Reg8], globals())

DelByVal = Instruction('DelByVal', 71, [Reg8, Reg8, Reg8], globals())

PutOwnGetterSetterByVal = Instruction(
    'PutOwnGetterSetterByVal', 72, [Reg8, Reg8, Reg8, Reg8, UInt8], globals()
)

GetPNameList = Instruction(
    'GetPNameList', 73, [Reg8, Reg8, Reg8, Reg8], globals()
)

GetNextPName = Instruction(
    'GetNextPName', 74, [Reg8, Reg8, Reg8, Reg8, Reg8], globals()
)

Call = Instruction('Call', 75, [Reg8, Reg8, UInt8], globals())
Call.has_ret_target = True

Construct = Instruction('Construct', 76, [Reg8, Reg8, UInt8], globals())
Construct.has_ret_target = True

Call1 = Instruction('Call1', 77, [Reg8, Reg8, Reg8], globals())
Call1.has_ret_target = True

CallDirect = Instruction('CallDirect', 78, [Reg8, UInt8, UInt16], globals())
CallDirect.has_ret_target = True

Call2 = Instruction('Call2', 79, [Reg8, Reg8, Reg8, Reg8], globals())
Call2.has_ret_target = True

Call3 = Instruction('Call3', 80, [Reg8, Reg8, Reg8, Reg8, Reg8], globals())
Call3.has_ret_target = True

Call4 = Instruction(
    'Call4', 81, [Reg8, Reg8, Reg8, Reg8, Reg8, Reg8], globals()
)
Call4.has_ret_target = True

CallLong = Instruction('CallLong', 82, [Reg8, Reg8, UInt32], globals())
CallLong.has_ret_target = True

ConstructLong = Instruction(
    'ConstructLong', 83, [Reg8, Reg8, UInt32], globals()
)
ConstructLong.has_ret_target = True

CallDirectLongIndex = Instruction(
    'CallDirectLongIndex', 84, [Reg8, UInt8, UInt32], globals()
)
CallDirectLongIndex.has_ret_target = True

CallBuiltin = Instruction('CallBuiltin', 85, [Reg8, UInt8, UInt8], globals())

CallBuiltinLong = Instruction(
    'CallBuiltinLong', 86, [Reg8, UInt8, UInt32], globals()
)

GetBuiltinClosure = Instruction(
    'GetBuiltinClosure', 87, [Reg8, UInt8], globals()
)

Ret = Instruction('Ret', 88, [Reg8], globals())

Catch = Instruction('Catch', 89, [Reg8], globals())

DirectEval = Instruction('DirectEval', 90, [Reg8, Reg8], globals())

Throw = Instruction('Throw', 91, [Reg8], globals())

ThrowIfEmpty = Instruction('ThrowIfEmpty', 92, [Reg8, Reg8], globals())

Debugger = Instruction('Debugger', 93, [], globals())

AsyncBreakCheck = Instruction('AsyncBreakCheck', 94, [], globals())

ProfilePoint = Instruction('ProfilePoint', 95, [UInt16], globals())

CreateClosure = Instruction(
    'CreateClosure', 96, [Reg8, Reg8, UInt16], globals()
)

CreateClosureLongIndex = Instruction(
    'CreateClosureLongIndex', 97, [Reg8, Reg8, UInt32], globals()
)

CreateGeneratorClosure = Instruction(
    'CreateGeneratorClosure', 98, [Reg8, Reg8, UInt16], globals()
)

CreateGeneratorClosureLongIndex = Instruction(
    'CreateGeneratorClosureLongIndex', 99, [Reg8, Reg8, UInt32], globals()
)

CreateAsyncClosure = Instruction(
    'CreateAsyncClosure', 100, [Reg8, Reg8, UInt16], globals()
)

CreateAsyncClosureLongIndex = Instruction(
    'CreateAsyncClosureLongIndex', 101, [Reg8, Reg8, UInt32], globals()
)

CreateThis = Instruction('CreateThis', 102, [Reg8, Reg8, Reg8], globals())

SelectObject = Instruction('SelectObject', 103, [Reg8, Reg8, Reg8], globals())

LoadParam = Instruction('LoadParam', 104, [Reg8, UInt8], globals())

LoadParamLong = Instruction('LoadParamLong', 105, [Reg8, UInt32], globals())

LoadConstUInt8 = Instruction('LoadConstUInt8', 106, [Reg8, UInt8], globals())

LoadConstInt = Instruction('LoadConstInt', 107, [Reg8, Imm32], globals())

LoadConstDouble = Instruction(
    'LoadConstDouble', 108, [Reg8, Double], globals()
)

LoadConstString = Instruction(
    'LoadConstString', 109, [Reg8, UInt16], globals()
)

LoadConstStringLongIndex = Instruction(
    'LoadConstStringLongIndex', 110, [Reg8, UInt32], globals()
)

LoadConstString.operands[1].operand_meaning = OperandMeaning.string_id

LoadConstStringLongIndex.operands[1].operand_meaning = OperandMeaning.string_id

LoadConstEmpty = Instruction('LoadConstEmpty', 111, [Reg8], globals())

LoadConstUndefined = Instruction('LoadConstUndefined', 112, [Reg8], globals())

LoadConstNull = Instruction('LoadConstNull', 113, [Reg8], globals())

LoadConstTrue = Instruction('LoadConstTrue', 114, [Reg8], globals())

LoadConstFalse = Instruction('LoadConstFalse', 115, [Reg8], globals())

LoadConstZero = Instruction('LoadConstZero', 116, [Reg8], globals())

CoerceThisNS = Instruction('CoerceThisNS', 117, [Reg8, Reg8], globals())

LoadThisNS = Instruction('LoadThisNS', 118, [Reg8], globals())

ToNumber = Instruction('ToNumber', 119, [Reg8, Reg8], globals())

ToInt32 = Instruction('ToInt32', 120, [Reg8, Reg8], globals())

AddEmptyString = Instruction('AddEmptyString', 121, [Reg8, Reg8], globals())

GetArgumentsPropByVal = Instruction(
    'GetArgumentsPropByVal', 122, [Reg8, Reg8, Reg8], globals()
)

GetArgumentsLength = Instruction(
    'GetArgumentsLength', 123, [Reg8, Reg8], globals()
)

ReifyArguments = Instruction('ReifyArguments', 124, [Reg8], globals())

CreateRegExp = Instruction(
    'CreateRegExp', 125, [Reg8, UInt32, UInt32, UInt32], globals()
)

CreateRegExp.operands[1].operand_meaning = OperandMeaning.string_id

CreateRegExp.operands[2].operand_meaning = OperandMeaning.string_id

SwitchImm = Instruction(
    'SwitchImm', 126, [Reg8, UInt32, Addr32, UInt32, UInt32], globals()
)

StartGenerator = Instruction('StartGenerator', 127, [], globals())

ResumeGenerator = Instruction('ResumeGenerator', 128, [Reg8, Reg8], globals())

CompleteGenerator = Instruction('CompleteGenerator', 129, [], globals())

CreateGenerator = Instruction(
    'CreateGenerator', 130, [Reg8, Reg8, UInt16], globals()
)

CreateGeneratorLongIndex = Instruction(
    'CreateGeneratorLongIndex', 131, [Reg8, Reg8, UInt32], globals()
)

IteratorBegin = Instruction('IteratorBegin', 132, [Reg8, Reg8], globals())

IteratorNext = Instruction('IteratorNext', 133, [Reg8, Reg8, Reg8], globals())

IteratorClose = Instruction('IteratorClose', 134, [Reg8, UInt8], globals())

Jmp = Instruction('Jmp', 135, [Addr8], globals())
JmpLong = Instruction('JmpLong', 136, [Addr32], globals())

JmpTrue = Instruction('JmpTrue', 137, [Addr8, Reg8], globals())
JmpTrueLong = Instruction('JmpTrueLong', 138, [Addr32, Reg8], globals())

JmpFalse = Instruction('JmpFalse', 139, [Addr8, Reg8], globals())
JmpFalseLong = Instruction('JmpFalseLong', 140, [Addr32, Reg8], globals())

JmpUndefined = Instruction('JmpUndefined', 141, [Addr8, Reg8], globals())
JmpUndefinedLong = Instruction(
    'JmpUndefinedLong', 142, [Addr32, Reg8], globals()
)

SaveGenerator = Instruction('SaveGenerator', 143, [Addr8], globals())
SaveGeneratorLong = Instruction('SaveGeneratorLong', 144, [Addr32], globals())

JLess = Instruction('JLess', 145, [Addr8, Reg8, Reg8], globals())
JLessLong = Instruction('JLessLong', 146, [Addr32, Reg8, Reg8], globals())

JNotLess = Instruction('JNotLess', 147, [Addr8, Reg8, Reg8], globals())
JNotLessLong = Instruction(
    'JNotLessLong', 148, [Addr32, Reg8, Reg8], globals()
)

JLessN = Instruction('JLessN', 149, [Addr8, Reg8, Reg8], globals())
JLessNLong = Instruction('JLessNLong', 150, [Addr32, Reg8, Reg8], globals())

JNotLessN = Instruction('JNotLessN', 151, [Addr8, Reg8, Reg8], globals())
JNotLessNLong = Instruction(
    'JNotLessNLong', 152, [Addr32, Reg8, Reg8], globals()
)

JLessEqual = Instruction('JLessEqual', 153, [Addr8, Reg8, Reg8], globals())
JLessEqualLong = Instruction(
    'JLessEqualLong', 154, [Addr32, Reg8, Reg8], globals()
)

JNotLessEqual = Instruction(
    'JNotLessEqual', 155, [Addr8, Reg8, Reg8], globals()
)
JNotLessEqualLong = Instruction(
    'JNotLessEqualLong', 156, [Addr32, Reg8, Reg8], globals()
)

JLessEqualN = Instruction('JLessEqualN', 157, [Addr8, Reg8, Reg8], globals())
JLessEqualNLong = Instruction(
    'JLessEqualNLong', 158, [Addr32, Reg8, Reg8], globals()
)

JNotLessEqualN = Instruction(
    'JNotLessEqualN', 159, [Addr8, Reg8, Reg8], globals()
)
JNotLessEqualNLong = Instruction(
    'JNotLessEqualNLong', 160, [Addr32, Reg8, Reg8], globals()
)

JGreater = Instruction('JGreater', 161, [Addr8, Reg8, Reg8], globals())
JGreaterLong = Instruction(
    'JGreaterLong', 162, [Addr32, Reg8, Reg8], globals()
)

JNotGreater = Instruction('JNotGreater', 163, [Addr8, Reg8, Reg8], globals())
JNotGreaterLong = Instruction(
    'JNotGreaterLong', 164, [Addr32, Reg8, Reg8], globals()
)

JGreaterN = Instruction('JGreaterN', 165, [Addr8, Reg8, Reg8], globals())
JGreaterNLong = Instruction(
    'JGreaterNLong', 166, [Addr32, Reg8, Reg8], globals()
)

JNotGreaterN = Instruction('JNotGreaterN', 167, [Addr8, Reg8, Reg8], globals())
JNotGreaterNLong = Instruction(
    'JNotGreaterNLong', 168, [Addr32, Reg8, Reg8], globals()
)

JGreaterEqual = Instruction(
    'JGreaterEqual', 169, [Addr8, Reg8, Reg8], globals()
)
JGreaterEqualLong = Instruction(
    'JGreaterEqualLong', 170, [Addr32, Reg8, Reg8], globals()
)

JNotGreaterEqual = Instruction(
    'JNotGreaterEqual', 171, [Addr8, Reg8, Reg8], globals()
)
JNotGreaterEqualLong = Instruction(
    'JNotGreaterEqualLong', 172, [Addr32, Reg8, Reg8], globals()
)

JGreaterEqualN = Instruction(
    'JGreaterEqualN', 173, [Addr8, Reg8, Reg8], globals()
)
JGreaterEqualNLong = Instruction(
    'JGreaterEqualNLong', 174, [Addr32, Reg8, Reg8], globals()
)

JNotGreaterEqualN = Instruction(
    'JNotGreaterEqualN', 175, [Addr8, Reg8, Reg8], globals()
)
JNotGreaterEqualNLong = Instruction(
    'JNotGreaterEqualNLong', 176, [Addr32, Reg8, Reg8], globals()
)

JEqual = Instruction('JEqual', 177, [Addr8, Reg8, Reg8], globals())
JEqualLong = Instruction('JEqualLong', 178, [Addr32, Reg8, Reg8], globals())

JNotEqual = Instruction('JNotEqual', 179, [Addr8, Reg8, Reg8], globals())
JNotEqualLong = Instruction(
    'JNotEqualLong', 180, [Addr32, Reg8, Reg8], globals()
)

JStrictEqual = Instruction('JStrictEqual', 181, [Addr8, Reg8, Reg8], globals())
JStrictEqualLong = Instruction(
    'JStrictEqualLong', 182, [Addr32, Reg8, Reg8], globals()
)

JStrictNotEqual = Instruction(
    'JStrictNotEqual', 183, [Addr8, Reg8, Reg8], globals()
)
JStrictNotEqualLong = Instruction(
    'JStrictNotEqualLong', 184, [Addr32, Reg8, Reg8], globals()
)

Add32 = Instruction('Add32', 185, [Reg8, Reg8, Reg8], globals())

Sub32 = Instruction('Sub32', 186, [Reg8, Reg8, Reg8], globals())

Mul32 = Instruction('Mul32', 187, [Reg8, Reg8, Reg8], globals())

Divi32 = Instruction('Divi32', 188, [Reg8, Reg8, Reg8], globals())

Divu32 = Instruction('Divu32', 189, [Reg8, Reg8, Reg8], globals())

Loadi8 = Instruction('Loadi8', 190, [Reg8, Reg8, Reg8], globals())

Loadu8 = Instruction('Loadu8', 191, [Reg8, Reg8, Reg8], globals())

Loadi16 = Instruction('Loadi16', 192, [Reg8, Reg8, Reg8], globals())

Loadu16 = Instruction('Loadu16', 193, [Reg8, Reg8, Reg8], globals())

Loadi32 = Instruction('Loadi32', 194, [Reg8, Reg8, Reg8], globals())

Loadu32 = Instruction('Loadu32', 195, [Reg8, Reg8, Reg8], globals())

Store8 = Instruction('Store8', 196, [Reg8, Reg8, Reg8], globals())

Store16 = Instruction('Store16', 197, [Reg8, Reg8, Reg8], globals())

Store32 = Instruction('Store32', 198, [Reg8, Reg8, Reg8], globals())

CallDirect.operands[2].operand_meaning = OperandMeaning.function_id

CreateClosure.operands[2].operand_meaning = OperandMeaning.function_id

CreateClosureLongIndex.operands[2].operand_meaning = OperandMeaning.function_id

CreateGeneratorClosure.operands[2].operand_meaning = OperandMeaning.function_id

CreateGeneratorClosureLongIndex.operands[
    2
].operand_meaning = OperandMeaning.function_id

CreateGenerator.operands[2].operand_meaning = OperandMeaning.function_id

CreateGeneratorLongIndex.operands[
    2
].operand_meaning = OperandMeaning.function_id

CreateAsyncClosure.operands[2].operand_meaning = OperandMeaning.function_id

CreateAsyncClosureLongIndex.operands[
    2
].operand_meaning = OperandMeaning.function_id

_opcode_to_instruction: Dict[int, Instruction] = {
    v.opcode: v for v in _instructions
}
_name_to_instruction: Dict[str, Instruction] = {
    v.name: v for v in _instructions
}

_builtin_function_names: List[str] = [
    'Array.isArray',
    'ArrayBuffer.isView',
    'Date.UTC',
    'Date.now',
    'Date.parse',
    'JSON.parse',
    'JSON.stringify',
    'Math.abs',
    'Math.acos',
    'Math.asin',
    'Math.atan',
    'Math.atan2',
    'Math.ceil',
    'Math.cos',
    'Math.exp',
    'Math.floor',
    'Math.hypot',
    'Math.imul',
    'Math.log',
    'Math.max',
    'Math.min',
    'Math.pow',
    'Math.random',
    'Math.round',
    'Math.sin',
    'Math.sqrt',
    'Math.tan',
    'Math.trunc',
    'Object.create',
    'Object.defineProperties',
    'Object.defineProperty',
    'Object.freeze',
    'Object.getOwnPropertyDescriptor',
    'Object.getOwnPropertyNames',
    'Object.getPrototypeOf',
    'Object.isExtensible',
    'Object.isFrozen',
    'Object.keys',
    'Object.seal',
    'String.fromCharCode',
    'silentSetPrototypeOf',
    'requireFast',
    'getTemplateObject',
    'ensureObject',
    'getMethod',
    'throwTypeError',
    'generatorSetDelegated',
    'copyDataProperties',
    'copyRestArgs',
    'arraySpread',
    'apply',
    'exportAll',
    'exponentiationOperator',
    'spawnAsync',
]



#!/usr/bin/env python3
"""
Note: The contents of the current file have been automatically
generated by the "utils/hermes_bytecode_structs_parser.py"
script

Please do not edit it manually. 👍
"""

from typing import List, Set, Dict, Union, Optional, Sequence, Any

# Imports from the current diretory
from .def_classes import *

_instructions: List[Instruction] = []

Reg8 = OperandType('Reg8', 'uint8_t')
Reg32 = OperandType('Reg32', 'uint32_t')
UInt8 = OperandType('UInt8', 'uint8_t')
UInt16 = OperandType('UInt16', 'uint16_t')
UInt32 = OperandType('UInt32', 'uint32_t')
Addr8 = OperandType('Addr8', 'int8_t')
Addr32 = OperandType('Addr32', 'int32_t')
Imm32 = OperandType('Imm32', 'int32_t')
Double = OperandType('Double', 'double')

Unreachable = Instruction('Unreachable', 0, [], globals())

NewObjectWithBuffer = Instruction(
    'NewObjectWithBuffer', 1, [Reg8, UInt16, UInt16], globals()
)

NewObjectWithBufferLong = Instruction(
    'NewObjectWithBufferLong', 2, [Reg8, UInt32, UInt32], globals()
)

NewObjectWithBufferAndParent = Instruction(
    'NewObjectWithBufferAndParent', 3, [Reg8, Reg8, UInt32, UInt32], globals()
)

NewObject = Instruction('NewObject', 4, [Reg8], globals())

NewObjectWithParent = Instruction(
    'NewObjectWithParent', 5, [Reg8, Reg8], globals()
)

NewArrayWithBuffer = Instruction(
    'NewArrayWithBuffer', 6, [Reg8, UInt16, UInt16, UInt16], globals()
)

NewArrayWithBufferLong = Instruction(
    'NewArrayWithBufferLong', 7, [Reg8, UInt16, UInt16, UInt32], globals()
)

NewArray = Instruction('NewArray', 8, [Reg8, UInt16], globals())

NewFastArray = Instruction('NewFastArray', 9, [Reg8, UInt16], globals())

FastArrayLength = Instruction('FastArrayLength', 10, [Reg8, Reg8], globals())

FastArrayLoad = Instruction('FastArrayLoad', 11, [Reg8, Reg8, Reg8], globals())

FastArrayStore = Instruction(
    'FastArrayStore', 12, [Reg8, Reg8, Reg8], globals()
)

FastArrayPush = Instruction('FastArrayPush', 13, [Reg8, Reg8], globals())

FastArrayAppend = Instruction('FastArrayAppend', 14, [Reg8, Reg8], globals())

CacheNewObject = Instruction(
    'CacheNewObject', 15, [Reg8, Reg8, UInt32, UInt8], globals()
)

Mov = Instruction('Mov', 16, [Reg8, Reg8], globals())

MovLong = Instruction('MovLong', 17, [Reg32, Reg32], globals())

Negate = Instruction('Negate', 18, [Reg8, Reg8], globals())

Not = Instruction('Not', 19, [Reg8, Reg8], globals())

BitNot = Instruction('BitNot', 20, [Reg8, Reg8], globals())

TypeOf = Instruction('TypeOf', 21, [Reg8, Reg8], globals())

Eq = Instruction('Eq', 22, [Reg8, Reg8, Reg8], globals())

StrictEq = Instruction('StrictEq', 23, [Reg8, Reg8, Reg8], globals())

Neq = Instruction('Neq', 24, [Reg8, Reg8, Reg8], globals())

StrictNeq = Instruction('StrictNeq', 25, [Reg8, Reg8, Reg8], globals())

Less = Instruction('Less', 26, [Reg8, Reg8, Reg8], globals())

LessEq = Instruction('LessEq', 27, [Reg8, Reg8, Reg8], globals())

Greater = Instruction('Greater', 28, [Reg8, Reg8, Reg8], globals())

GreaterEq = Instruction('GreaterEq', 29, [Reg8, Reg8, Reg8], globals())

Add = Instruction('Add', 30, [Reg8, Reg8, Reg8], globals())

AddN = Instruction('AddN', 31, [Reg8, Reg8, Reg8], globals())

AddS = Instruction('AddS', 32, [Reg8, Reg8, Reg8], globals())

Mul = Instruction('Mul', 33, [Reg8, Reg8, Reg8], globals())

MulN = Instruction('MulN', 34, [Reg8, Reg8, Reg8], globals())

Div = Instruction('Div', 35, [Reg8, Reg8, Reg8], globals())

DivN = Instruction('DivN', 36, [Reg8, Reg8, Reg8], globals())

Mod = Instruction('Mod', 37, [Reg8, Reg8, Reg8], globals())

Sub = Instruction('Sub', 38, [Reg8, Reg8, Reg8], globals())

SubN = Instruction('SubN', 39, [Reg8, Reg8, Reg8], globals())

LShift = Instruction('LShift', 40, [Reg8, Reg8, Reg8], globals())

RShift = Instruction('RShift', 41, [Reg8, Reg8, Reg8], globals())

URshift = Instruction('URshift', 42, [Reg8, Reg8, Reg8], globals())

BitAnd = Instruction('BitAnd', 43, [Reg8, Reg8, Reg8], globals())

BitXor = Instruction('BitXor', 44, [Reg8, Reg8, Reg8], globals())

BitOr = Instruction('BitOr', 45, [Reg8, Reg8, Reg8], globals())

Inc = Instruction('Inc', 46, [Reg8, Reg8], globals())

Dec = Instruction('Dec', 47, [Reg8, Reg8], globals())

InstanceOf = Instruction('InstanceOf', 48, [Reg8, Reg8, Reg8], globals())

IsIn = Instruction('IsIn', 49, [Reg8, Reg8, Reg8], globals())

PrivateIsIn = Instruction(
    'PrivateIsIn', 50, [Reg8, Reg8, Reg8, Reg8], globals()
)

TypeOfIs = Instruction('TypeOfIs', 51, [Reg8, Reg8, UInt16], globals())

GetParentEnvironment = Instruction(
    'GetParentEnvironment', 52, [Reg8, UInt8], globals()
)

GetEnvironment = Instruction(
    'GetEnvironment', 53, [Reg8, Reg8, UInt8], globals()
)

GetClosureEnvironment = Instruction(
    'GetClosureEnvironment', 54, [Reg8, Reg8], globals()
)

StoreToEnvironment = Instruction(
    'StoreToEnvironment', 55, [Reg8, UInt8, Reg8], globals()
)

StoreToEnvironmentL = Instruction(
    'StoreToEnvironmentL', 56, [Reg8, UInt16, Reg8], globals()
)

StoreNPToEnvironment = Instruction(
    'StoreNPToEnvironment', 57, [Reg8, UInt8, Reg8], globals()
)

StoreNPToEnvironmentL = Instruction(
    'StoreNPToEnvironmentL', 58, [Reg8, UInt16, Reg8], globals()
)

LoadFromEnvironment = Instruction(
    'LoadFromEnvironment', 59, [Reg8, Reg8, UInt8], globals()
)

LoadFromEnvironmentL = Instruction(
    'LoadFromEnvironmentL', 60, [Reg8, Reg8, UInt16], globals()
)

GetGlobalObject = Instruction('GetGlobalObject', 61, [Reg8], globals())

GetNewTarget = Instruction('GetNewTarget', 62, [Reg8], globals())

LoadParentNoTraps = Instruction(
    'LoadParentNoTraps', 63, [Reg8, Reg8], globals()
)

CreateFunctionEnvironment = Instruction(
    'CreateFunctionEnvironment', 64, [Reg8, UInt8], globals()
)

CreateTopLevelEnvironment = Instruction(
    'CreateTopLevelEnvironment', 65, [Reg8, UInt32], globals()
)

CreateEnvironment = Instruction(
    'CreateEnvironment', 66, [Reg8, Reg8, UInt32], globals()
)

DeclareGlobalVar = Instruction('DeclareGlobalVar', 67, [UInt32], globals())

DeclareGlobalVar.operands[0].operand_meaning = OperandMeaning.string_id

GetByIdShort = Instruction(
    'GetByIdShort', 68, [Reg8, Reg8, UInt8, UInt8], globals()
)

GetById = Instruction('GetById', 69, [Reg8, Reg8, UInt8, UInt16], globals())

GetByIdLong = Instruction(
    'GetByIdLong', 70, [Reg8, Reg8, UInt8, UInt32], globals()
)

GetByIdShort.operands[3].operand_meaning = OperandMeaning.string_id

GetById.operands[3].operand_meaning = OperandMeaning.string_id

GetByIdLong.operands[3].operand_meaning = OperandMeaning.string_id

GetByIdWithReceiverLong = Instruction(
    'GetByIdWithReceiverLong', 71, [Reg8, Reg8, UInt8, Reg8, UInt32], globals()
)

GetByIdWithReceiverLong.operands[4].operand_meaning = OperandMeaning.string_id

TryGetById = Instruction(
    'TryGetById', 72, [Reg8, Reg8, UInt8, UInt16], globals()
)

TryGetByIdLong = Instruction(
    'TryGetByIdLong', 73, [Reg8, Reg8, UInt8, UInt32], globals()
)

TryGetById.operands[3].operand_meaning = OperandMeaning.string_id

TryGetByIdLong.operands[3].operand_meaning = OperandMeaning.string_id

PutByIdLoose = Instruction(
    'PutByIdLoose', 74, [Reg8, Reg8, UInt8, UInt16], globals()
)

PutByIdStrict = Instruction(
    'PutByIdStrict', 75, [Reg8, Reg8, UInt8, UInt16], globals()
)

PutByIdLooseLong = Instruction(
    'PutByIdLooseLong', 76, [Reg8, Reg8, UInt8, UInt32], globals()
)

PutByIdStrictLong = Instruction(
    'PutByIdStrictLong', 77, [Reg8, Reg8, UInt8, UInt32], globals()
)

PutByIdLoose.operands[3].operand_meaning = OperandMeaning.string_id

PutByIdStrict.operands[3].operand_meaning = OperandMeaning.string_id

PutByIdLooseLong.operands[3].operand_meaning = OperandMeaning.string_id

PutByIdStrictLong.operands[3].operand_meaning = OperandMeaning.string_id

TryPutByIdLoose = Instruction(
    'TryPutByIdLoose', 78, [Reg8, Reg8, UInt8, UInt16], globals()
)

TryPutByIdStrict = Instruction(
    'TryPutByIdStrict', 79, [Reg8, Reg8, UInt8, UInt16], globals()
)

TryPutByIdLooseLong = Instruction(
    'TryPutByIdLooseLong', 80, [Reg8, Reg8, UInt8, UInt32], globals()
)

TryPutByIdStrictLong = Instruction(
    'TryPutByIdStrictLong', 81, [Reg8, Reg8, UInt8, UInt32], globals()
)

TryPutByIdLoose.operands[3].operand_meaning = OperandMeaning.string_id

TryPutByIdStrict.operands[3].operand_meaning = OperandMeaning.string_id

TryPutByIdLooseLong.operands[3].operand_meaning = OperandMeaning.string_id

TryPutByIdStrictLong.operands[3].operand_meaning = OperandMeaning.string_id

PutOwnBySlotIdx = Instruction(
    'PutOwnBySlotIdx', 82, [Reg8, Reg8, UInt8], globals()
)

PutOwnBySlotIdxLong = Instruction(
    'PutOwnBySlotIdxLong', 83, [Reg8, Reg8, UInt32], globals()
)

GetOwnBySlotIdx = Instruction(
    'GetOwnBySlotIdx', 84, [Reg8, Reg8, UInt8], globals()
)

GetOwnBySlotIdxLong = Instruction(
    'GetOwnBySlotIdxLong', 85, [Reg8, Reg8, UInt32], globals()
)

DefineOwnById = Instruction(
    'DefineOwnById', 86, [Reg8, Reg8, UInt8, UInt16], globals()
)

DefineOwnByIdLong = Instruction(
    'DefineOwnByIdLong', 87, [Reg8, Reg8, UInt8, UInt32], globals()
)

DefineOwnByIdLong.operands[3].operand_meaning = OperandMeaning.string_id

DefineOwnByIndex = Instruction(
    'DefineOwnByIndex', 88, [Reg8, Reg8, UInt8], globals()
)

DefineOwnByIndexL = Instruction(
    'DefineOwnByIndexL', 89, [Reg8, Reg8, UInt32], globals()
)

DefineOwnInDenseArray = Instruction(
    'DefineOwnInDenseArray', 90, [Reg8, Reg8, UInt8], globals()
)

DefineOwnInDenseArrayL = Instruction(
    'DefineOwnInDenseArrayL', 91, [Reg8, Reg8, UInt16], globals()
)

DefineOwnByVal = Instruction(
    'DefineOwnByVal', 92, [Reg8, Reg8, Reg8, UInt8], globals()
)

GetByVal = Instruction('GetByVal', 93, [Reg8, Reg8, Reg8], globals())

GetByIndex = Instruction('GetByIndex', 94, [Reg8, Reg8, UInt8], globals())

GetByValWithReceiver = Instruction(
    'GetByValWithReceiver', 95, [Reg8, Reg8, Reg8, Reg8], globals()
)

PutByValLoose = Instruction('PutByValLoose', 96, [Reg8, Reg8, Reg8], globals())

PutByValStrict = Instruction(
    'PutByValStrict', 97, [Reg8, Reg8, Reg8], globals()
)

PutByValWithReceiver = Instruction(
    'PutByValWithReceiver', 98, [Reg8, Reg8, Reg8, Reg8, UInt8], globals()
)

DelByVal = Instruction('DelByVal', 99, [Reg8, Reg8, Reg8, UInt8], globals())

AddOwnPrivateBySym = Instruction(
    'AddOwnPrivateBySym', 100, [Reg8, Reg8, Reg8], globals()
)

GetOwnPrivateBySym = Instruction(
    'GetOwnPrivateBySym', 101, [Reg8, Reg8, UInt8, Reg8], globals()
)

PutOwnPrivateBySym = Instruction(
    'PutOwnPrivateBySym', 102, [Reg8, Reg8, UInt8, Reg8], globals()
)

DefineOwnGetterSetterByVal = Instruction(
    'DefineOwnGetterSetterByVal',
    103,
    [Reg8, Reg8, Reg8, Reg8, UInt8],
    globals(),
)

GetPNameList = Instruction(
    'GetPNameList', 104, [Reg8, Reg8, Reg8, Reg8], globals()
)

GetNextPName = Instruction(
    'GetNextPName', 105, [Reg8, Reg8, Reg8, Reg8, Reg8], globals()
)

Call = Instruction('Call', 106, [Reg8, Reg8, UInt8], globals())
Call.has_ret_target = True

Construct = Instruction('Construct', 107, [Reg8, Reg8, UInt8], globals())
Construct.has_ret_target = True

Call1 = Instruction('Call1', 108, [Reg8, Reg8, Reg8], globals())
Call1.has_ret_target = True

CallWithNewTarget = Instruction(
    'CallWithNewTarget', 109, [Reg8, Reg8, Reg8, UInt8], globals()
)
CallWithNewTarget.has_ret_target = True

Call2 = Instruction('Call2', 110, [Reg8, Reg8, Reg8, Reg8], globals())
Call2.has_ret_target = True

Call3 = Instruction('Call3', 111, [Reg8, Reg8, Reg8, Reg8, Reg8], globals())
Call3.has_ret_target = True

Call4 = Instruction(
    'Call4', 112, [Reg8, Reg8, Reg8, Reg8, Reg8, Reg8], globals()
)
Call4.has_ret_target = True

CallWithNewTargetLong = Instruction(
    'CallWithNewTargetLong', 113, [Reg8, Reg8, Reg8, Reg8], globals()
)
CallWithNewTargetLong.has_ret_target = True

CallRequire = Instruction('CallRequire', 114, [Reg8, Reg8, UInt32], globals())
CallRequire.has_ret_target = True

CallBuiltin = Instruction('CallBuiltin', 115, [Reg8, UInt8, UInt8], globals())

CallBuiltinLong = Instruction(
    'CallBuiltinLong', 116, [Reg8, UInt8, UInt32], globals()
)

GetBuiltinClosure = Instruction(
    'GetBuiltinClosure', 117, [Reg8, UInt8], globals()
)

Ret = Instruction('Ret', 118, [Reg8], globals())

Catch = Instruction('Catch', 119, [Reg8], globals())

DirectEval = Instruction('DirectEval', 120, [Reg8, Reg8, UInt8], globals())

Throw = Instruction('Throw', 121, [Reg8], globals())

ThrowIfEmpty = Instruction('ThrowIfEmpty', 122, [Reg8, Reg8], globals())

ThrowIfUndefined = Instruction(
    'ThrowIfUndefined', 123, [Reg8, Reg8], globals()
)

ThrowIfThisInitialized = Instruction(
    'ThrowIfThisInitialized', 124, [Reg8], globals()
)

Debugger = Instruction('Debugger', 125, [], globals())

AsyncBreakCheck = Instruction('AsyncBreakCheck', 126, [], globals())

ProfilePoint = Instruction('ProfilePoint', 127, [UInt16], globals())

CreateBaseClass = Instruction(
    'CreateBaseClass', 128, [Reg8, Reg8, Reg8, UInt16], globals()
)

CreateBaseClassLongIndex = Instruction(
    'CreateBaseClassLongIndex', 129, [Reg8, Reg8, Reg8, UInt32], globals()
)

CreateDerivedClass = Instruction(
    'CreateDerivedClass', 130, [Reg8, Reg8, Reg8, Reg8, UInt16], globals()
)

CreateDerivedClassLongIndex = Instruction(
    'CreateDerivedClassLongIndex',
    131,
    [Reg8, Reg8, Reg8, Reg8, UInt32],
    globals(),
)

CreateClosure = Instruction(
    'CreateClosure', 132, [Reg8, Reg8, UInt16], globals()
)

CreateClosureLongIndex = Instruction(
    'CreateClosureLongIndex', 133, [Reg8, Reg8, UInt32], globals()
)

CreateClosure.operands[2].operand_meaning = OperandMeaning.function_id

CreateClosureLongIndex.operands[2].operand_meaning = OperandMeaning.function_id

CreateThisForNew = Instruction(
    'CreateThisForNew', 134, [Reg8, Reg8, UInt8], globals()
)

CreateThisForSuper = Instruction(
    'CreateThisForSuper', 135, [Reg8, Reg8, Reg8, UInt8], globals()
)

SelectObject = Instruction('SelectObject', 136, [Reg8, Reg8, Reg8], globals())

LoadParam = Instruction('LoadParam', 137, [Reg8, UInt8], globals())

LoadParamLong = Instruction('LoadParamLong', 138, [Reg8, UInt32], globals())

LoadConstUInt8 = Instruction('LoadConstUInt8', 139, [Reg8, UInt8], globals())

LoadConstInt = Instruction('LoadConstInt', 140, [Reg8, Imm32], globals())

LoadConstDouble = Instruction(
    'LoadConstDouble', 141, [Reg8, Double], globals()
)

LoadConstBigInt = Instruction(
    'LoadConstBigInt', 142, [Reg8, UInt16], globals()
)

LoadConstBigIntLongIndex = Instruction(
    'LoadConstBigIntLongIndex', 143, [Reg8, UInt32], globals()
)

LoadConstBigInt.operands[1].operand_meaning = OperandMeaning.bigint_id

LoadConstBigIntLongIndex.operands[1].operand_meaning = OperandMeaning.bigint_id

LoadConstString = Instruction(
    'LoadConstString', 144, [Reg8, UInt16], globals()
)

LoadConstStringLongIndex = Instruction(
    'LoadConstStringLongIndex', 145, [Reg8, UInt32], globals()
)

LoadConstString.operands[1].operand_meaning = OperandMeaning.string_id

LoadConstStringLongIndex.operands[1].operand_meaning = OperandMeaning.string_id

LoadConstEmpty = Instruction('LoadConstEmpty', 146, [Reg8], globals())

LoadConstUndefined = Instruction('LoadConstUndefined', 147, [Reg8], globals())

LoadConstNull = Instruction('LoadConstNull', 148, [Reg8], globals())

LoadConstTrue = Instruction('LoadConstTrue', 149, [Reg8], globals())

LoadConstFalse = Instruction('LoadConstFalse', 150, [Reg8], globals())

LoadConstZero = Instruction('LoadConstZero', 151, [Reg8], globals())

CoerceThisNS = Instruction('CoerceThisNS', 152, [Reg8, Reg8], globals())

LoadThisNS = Instruction('LoadThisNS', 153, [Reg8], globals())

ToNumber = Instruction('ToNumber', 154, [Reg8, Reg8], globals())

ToNumeric = Instruction('ToNumeric', 155, [Reg8, Reg8], globals())

ToInt32 = Instruction('ToInt32', 156, [Reg8, Reg8], globals())

AddEmptyString = Instruction('AddEmptyString', 157, [Reg8, Reg8], globals())

CreatePrivateName = Instruction(
    'CreatePrivateName', 158, [Reg8, UInt32], globals()
)

CreatePrivateName.operands[1].operand_meaning = OperandMeaning.string_id

GetArgumentsPropByValLoose = Instruction(
    'GetArgumentsPropByValLoose', 159, [Reg8, Reg8, Reg8], globals()
)

GetArgumentsPropByValStrict = Instruction(
    'GetArgumentsPropByValStrict', 160, [Reg8, Reg8, Reg8], globals()
)

GetArgumentsLength = Instruction(
    'GetArgumentsLength', 161, [Reg8, Reg8], globals()
)

ReifyArgumentsStrict = Instruction(
    'ReifyArgumentsStrict', 162, [Reg8], globals()
)

ReifyArgumentsLoose = Instruction(
    'ReifyArgumentsLoose', 163, [Reg8], globals()
)

ToPropertyKey = Instruction('ToPropertyKey', 164, [Reg8, Reg8], globals())

CreateRegExp = Instruction(
    'CreateRegExp', 165, [Reg8, UInt32, UInt32, UInt32], globals()
)

CreateRegExp.operands[1].operand_meaning = OperandMeaning.string_id

CreateRegExp.operands[2].operand_meaning = OperandMeaning.string_id

UIntSwitchImm = Instruction(
    'UIntSwitchImm', 166, [Reg8, UInt32, Addr32, UInt32, UInt32], globals()
)

StringSwitchImm = Instruction(
    'StringSwitchImm', 167, [Reg8, UInt32, UInt32, Addr32, UInt32], globals()
)

CreateGenerator = Instruction(
    'CreateGenerator', 168, [Reg8, Reg8, UInt16], globals()
)

CreateGeneratorLongIndex = Instruction(
    'CreateGeneratorLongIndex', 169, [Reg8, Reg8, UInt32], globals()
)

CreateGenerator.operands[2].operand_meaning = OperandMeaning.function_id

CreateGeneratorLongIndex.operands[
    2
].operand_meaning = OperandMeaning.function_id

IteratorBegin = Instruction('IteratorBegin', 170, [Reg8, Reg8], globals())

IteratorNext = Instruction('IteratorNext', 171, [Reg8, Reg8, Reg8], globals())

IteratorClose = Instruction('IteratorClose', 172, [Reg8, UInt8], globals())

TypedLoadParent = Instruction('TypedLoadParent', 173, [Reg8, Reg8], globals())

Jmp = Instruction('Jmp', 174, [Addr8], globals())
JmpLong = Instruction('JmpLong', 175, [Addr32], globals())

JmpTrue = Instruction('JmpTrue', 176, [Addr8, Reg8], globals())
JmpTrueLong = Instruction('JmpTrueLong', 177, [Addr32, Reg8], globals())

JmpFalse = Instruction('JmpFalse', 178, [Addr8, Reg8], globals())
JmpFalseLong = Instruction('JmpFalseLong', 179, [Addr32, Reg8], globals())

JmpUndefined = Instruction('JmpUndefined', 180, [Addr8, Reg8], globals())
JmpUndefinedLong = Instruction(
    'JmpUndefinedLong', 181, [Addr32, Reg8], globals()
)

JmpTypeOfIs = Instruction(
    'JmpTypeOfIs', 182, [Addr32, Reg8, UInt16], globals()
)

JLess = Instruction('JLess', 183, [Addr8, Reg8, Reg8], globals())
JLessLong = Instruction('JLessLong', 184, [Addr32, Reg8, Reg8], globals())

JNotLess = Instruction('JNotLess', 185, [Addr8, Reg8, Reg8], globals())
JNotLessLong = Instruction(
    'JNotLessLong', 186, [Addr32, Reg8, Reg8], globals()
)

JLessN = Instruction('JLessN', 187, [Addr8, Reg8, Reg8], globals())
JLessNLong = Instruction('JLessNLong', 188, [Addr32, Reg8, Reg8], globals())

JNotLessN = Instruction('JNotLessN', 189, [Addr8, Reg8, Reg8], globals())
JNotLessNLong = Instruction(
    'JNotLessNLong', 190, [Addr32, Reg8, Reg8], globals()
)

JLessEqual = Instruction('JLessEqual', 191, [Addr8, Reg8, Reg8], globals())
JLessEqualLong = Instruction(
    'JLessEqualLong', 192, [Addr32, Reg8, Reg8], globals()
)

JNotLessEqual = Instruction(
    'JNotLessEqual', 193, [Addr8, Reg8, Reg8], globals()
)
JNotLessEqualLong = Instruction(
    'JNotLessEqualLong', 194, [Addr32, Reg8, Reg8], globals()
)

JLessEqualN = Instruction('JLessEqualN', 195, [Addr8, Reg8, Reg8], globals())
JLessEqualNLong = Instruction(
    'JLessEqualNLong', 196, [Addr32, Reg8, Reg8], globals()
)

JNotLessEqualN = Instruction(
    'JNotLessEqualN', 197, [Addr8, Reg8, Reg8], globals()
)
JNotLessEqualNLong = Instruction(
    'JNotLessEqualNLong', 198, [Addr32, Reg8, Reg8], globals()
)

JGreater = Instruction('JGreater', 199, [Addr8, Reg8, Reg8], globals())
JGreaterLong = Instruction(
    'JGreaterLong', 200, [Addr32, Reg8, Reg8], globals()
)

JNotGreater = Instruction('JNotGreater', 201, [Addr8, Reg8, Reg8], globals())
JNotGreaterLong = Instruction(
    'JNotGreaterLong', 202, [Addr32, Reg8, Reg8], globals()
)

JGreaterEqual = Instruction(
    'JGreaterEqual', 203, [Addr8, Reg8, Reg8], globals()
)
JGreaterEqualLong = Instruction(
    'JGreaterEqualLong', 204, [Addr32, Reg8, Reg8], globals()
)

JNotGreaterEqual = Instruction(
    'JNotGreaterEqual', 205, [Addr8, Reg8, Reg8], globals()
)
JNotGreaterEqualLong = Instruction(
    'JNotGreaterEqualLong', 206, [Addr32, Reg8, Reg8], globals()
)

JEqual = Instruction('JEqual', 207, [Addr8, Reg8, Reg8], globals())
JEqualLong = Instruction('JEqualLong', 208, [Addr32, Reg8, Reg8], globals())

JNotEqual = Instruction('JNotEqual', 209, [Addr8, Reg8, Reg8], globals())
JNotEqualLong = Instruction(
    'JNotEqualLong', 210, [Addr32, Reg8, Reg8], globals()
)

JStrictEqual = Instruction('JStrictEqual', 211, [Addr8, Reg8, Reg8], globals())
JStrictEqualLong = Instruction(
    'JStrictEqualLong', 212, [Addr32, Reg8, Reg8], globals()
)

JStrictNotEqual = Instruction(
    'JStrictNotEqual', 213, [Addr8, Reg8, Reg8], globals()
)
JStrictNotEqualLong = Instruction(
    'JStrictNotEqualLong', 214, [Addr32, Reg8, Reg8], globals()
)

JmpBuiltinIs = Instruction(
    'JmpBuiltinIs', 215, [Addr8, UInt8, Reg8], globals()
)

JmpBuiltinIsLong = Instruction(
    'JmpBuiltinIsLong', 216, [Addr32, UInt8, Reg8], globals()
)

JmpBuiltinIsNot = Instruction(
    'JmpBuiltinIsNot', 217, [Addr8, UInt8, Reg8], globals()
)

JmpBuiltinIsNotLong = Instruction(
    'JmpBuiltinIsNotLong', 218, [Addr32, UInt8, Reg8], globals()
)

_opcode_to_instruction: Dict[int, Instruction] = {
    v.opcode: v for v in _instructions
}
_name_to_instruction: Dict[str, Instruction] = {
    v.name: v for v in _instructions
}

_builtin_function_names: List[str] = [
    'Array.isArray',
    'Date.UTC',
    'Date.parse',
    'JSON.parse',
    'JSON.stringify',
    'Math.abs',
    'Math.acos',
    'Math.asin',
    'Math.atan',
    'Math.atan2',
    'Math.ceil',
    'Math.cos',
    'Math.exp',
    'Math.floor',
    'Math.hypot',
    'Math.imul',
    'Math.log',
    'Math.max',
    'Math.min',
    'Math.pow',
    'Math.round',
    'Math.sin',
    'Math.sqrt',
    'Math.tan',
    'Math.trunc',
    'Object.create',
    'Object.defineProperties',
    'Object.defineProperty',
    'Object.freeze',
    'Object.getOwnPropertyDescriptor',
    'Object.getOwnPropertyNames',
    'Object.getPrototypeOf',
    'Object.isExtensible',
    'Object.isFrozen',
    'Object.keys',
    'Object.seal',
    'String.fromCharCode',
    'silentSetPrototypeOf',
    'requireFast',
    'getTemplateObject',
    'ensureObject',
    'getMethod',
    'throwTypeError',
    'throwReferenceError',
    'copyDataProperties',
    'copyRestArgs',
    'arraySpread',
    'apply',
    'applyArguments',
    'applyWithNewTarget',
    'exportAll',
    'exponentiationOperator',
    'initRegexNamedGroups',
    'functionPrototypeApply',
    'functionPrototypeCall',
    'spawnAsync',
    'makeAsyncIterator',
]



#!/usr/bin/env python3
# -*- encoding: Utf-8 -*-
from typing import List, Dict, Set, Sequence, Union, Optional, Iterator, Any
from logging import warning
from io import BytesIO

from hermes_dec.parsers.hbc_opcodes import (
    hbc51,
    hbc58,
    hbc59,
    hbc61,
    hbc62,
    hbc68,
    hbc69,
    hbc70,
    hbc72,
    hbc73,
    hbc76,
    hbc80,
    hbc81,
    hbc82,
    hbc83,
    hbc84,
    hbc85,
    hbc86,
    hbc87,
    hbc89,
    hbc90,
    hbc92,
    hbc95,
    hbc97,
    hbc98,
    hbc99,
)
from hermes_dec.parsers.serialized_literal_parser import (
    unpack_slp_array,
    SLPArray,
    SLPValue,
    TagType,
)
from hermes_dec.parsers.hbc_opcodes.def_classes import (
    OperandMeaning,
    Instruction,
)


class ParsedInstruction:
    inst: Instruction
    arg1: object
    arg2: object
    arg3: object
    arg4: object
    arg5: object
    arg6: object
    switch_jump_table: Optional[List[int]]
    original_pos: int
    next_pos: int
    hbc_reader: 'HBCReader'

    def __repr__(self):
        operands = [
            '%s: %s'
            % (
                (
                    self.inst.operands[index].operand_meaning.name
                    if self.inst.operands[index].operand_meaning
                    else self.inst.operands[index].operand_type.name
                ),
                getattr(self, 'arg%d' % (index + 1)),
            )
            for index in range(len(self.inst.operands))
        ]

        comment = ''
        for operand_index, operand in enumerate(self.inst.operands):
            if operand.operand_meaning:
                operand_value = getattr(self, 'arg%d' % (operand_index + 1))
                if operand.operand_meaning == OperandMeaning.string_id:
                    comment += '  # String: %r (%s)' % (
                        self.hbc_reader.strings[operand_value],
                        self.hbc_reader.string_kinds[operand_value].name,
                    )
                elif operand.operand_meaning == OperandMeaning.bigint_id:
                    comment += (
                        '  # BigInt: %s'
                        % (self.hbc_reader.bigint_values[operand_value])
                    )
                elif operand.operand_meaning == OperandMeaning.function_id:
                    function_header = self.hbc_reader.function_headers[
                        operand_value
                    ]
                    comment += (
                        '  # Function: [#%d %s of %d bytes]: %d params @ offset 0x%08x'
                        % (
                            operand_value,
                            self.hbc_reader.strings[
                                function_header.functionName
                            ],
                            function_header.bytecodeSizeInBytes,
                            function_header.paramCount,
                            function_header.offset,
                        )
                    )
            elif operand.operand_type.name in ('Addr8', 'Addr32'):
                operand_value = getattr(self, 'arg%d' % (operand_index + 1))
                comment += '  # Address: %08x' % (
                    self.original_pos + operand_value
                )
        if self.inst.name in ('NewArrayWithBuffer', 'NewArrayWithBufferLong'):
            comment += '  # Array: [%s]' % ', '.join(
                unpack_slp_array(
                    (
                        self.hbc_reader.arrays
                        if self.hbc_reader.header.version < 97
                        else self.hbc_reader.literal_values
                    )[self.arg4 :],
                    self.arg3,
                ).to_strings(self.hbc_reader.strings)
            )
        elif self.inst.name in (
            'NewObjectWithBuffer',
            'NewObjectWithBufferLong',
        ):
            if self.hbc_reader.header.version < 97:
                comment += '  # Object: {%s}' % ', '.join(
                    '%s: %s' % (key, value)
                    for key, value in zip(
                        unpack_slp_array(
                            self.hbc_reader.object_keys[self.arg4 :], self.arg3
                        ).to_strings(self.hbc_reader.strings),
                        unpack_slp_array(
                            (self.hbc_reader.object_values)[self.arg5 :],
                            self.arg3,
                        ).to_strings(self.hbc_reader.strings),
                    )
                )
            else:
                shape_keys = self.hbc_reader.object_shape_keys[self.arg2]
                comment += '  # Object: {%s}' % ', '.join(
                    '%s: %s' % (key, value)
                    for key, value in zip(
                        shape_keys,
                        unpack_slp_array(
                            (self.hbc_reader.literal_values)[self.arg3 :],
                            len(shape_keys),
                        ).to_strings(self.hbc_reader.strings),
                    )
                )
        elif self.inst.name in (
            'CallBuiltin',
            'CallBuiltinLong',
            'GetBuiltinClosure',
        ):
            builtin_number = self.arg2
            builtin_functions = get_builtin_functions(
                self.hbc_reader.parser_module
            )
            comment += '  # Built-in function: [#%d %s]' % (
                builtin_number,
                builtin_functions[builtin_number],
            )
        elif self.inst.name == 'SwitchImm':
            comment += '  # Jump table: [%s]' % ', '.join(
                '%08x' % value for value in self.switch_jump_table
            )

        return f'{"%08x" % self.original_pos}: <{self.inst.name}>: <{", ".join(operands)}>{comment}'


def get_parser(bytecode_version: int) -> 'module':
    parser_module_tbl = {
        51: hbc51,
        58: hbc58,
        59: hbc59,
        61: hbc61,
        62: hbc62,
        68: hbc68,
        69: hbc69,
        70: hbc70,
        # Before version 72/0.4.0, the bytecode format used predefined
        # string identifiers (which were removed in commit 63fb517)
        # These are not supported in the current code setup
        72: hbc72,
        73: hbc73,
        76: hbc76,
        80: hbc80,
        81: hbc81,
        82: hbc82,
        83: hbc83,
        84: hbc84,
        85: hbc85,
        86: hbc86,
        87: hbc87,
        89: hbc89,
        90: hbc90,
        92: hbc92,
        # The changes introduced in Hermes bytecode version 92
        # were reverted to the state of version 91 after a few
        # days from their introduction in the Git tree, into a
        # new version 93, and reintroduced in version 94.
        93: hbc90,
        94: hbc92,
        95: hbc95,
        # Bytecode version 96 has no opcode changes vs 95, only the
        # regex engine gains an INDICES flag (already handled).
        96: hbc95,
        # Bytecode version 97 marks the first updates in static Hermes
        # and is based off bytecode version 90
        97: hbc97,
        98: hbc98,
        99: hbc99,
    }

    if bytecode_version < 72:
        warning(
            'This file uses an ancient Hermes bytecode format, which '
            + 'is not supported.'
        )

    elif bytecode_version > 96:
        warning(
            (
                'Bytecode version %d corresponds to a development or '
                + 'recent version of the Hermes bytecode and is not '
                + 'formally supported by the current tool.'
            )
            % bytecode_version
        )

    for min_version in reversed(sorted(parser_module_tbl)):
        if bytecode_version >= min_version:
            parser_module = parser_module_tbl[min_version]
            break

    return parser_module


def get_builtin_functions(parser_module: 'module') -> List[str]:

    return parser_module._builtin_function_names


# Read and parse the actual bytecode (+ switch tables) for
# a given function header


def parse_hbc_bytecode(
    function_header: object, hbc_reader: 'HBCReader'
) -> Iterator[ParsedInstruction]:

    hbc_reader.file_buffer.seek(function_header.offset)

    buf = BytesIO(
        hbc_reader.file_buffer.read(function_header.bytecodeSizeInBytes)
    )

    while True:
        original_pos = buf.tell()

        opcode = buf.read(1)
        if not opcode:
            break
        opcode = opcode[0]

        inst = hbc_reader.parser_module._opcode_to_instruction.get(opcode)
        if not inst:
            raise NotImplementedError(
                'Opcode "0x%02x" is currently not implemented within hermes-dec.'
                % opcode
            )

        structure = inst.structure()
        buf.readinto(structure)

        result = ParsedInstruction()
        result.inst = inst
        result.original_pos = original_pos
        result.next_pos = original_pos + inst.binary_size
        result.hbc_reader = hbc_reader

        for operand in ('arg1', 'arg2', 'arg3', 'arg4', 'arg5', 'arg6'):
            if hasattr(structure, operand):
                setattr(result, operand, getattr(structure, operand))

        if inst.name == 'SwitchImm':
            result.switch_jump_table = []
            hbc_reader.file_buffer.seek(
                function_header.offset + original_pos + structure.arg2
            )
            hbc_reader.align_over_padding()

            for jump_table_entry in range(structure.arg4, structure.arg5 + 1):
                result.switch_jump_table.append(
                    int.from_bytes(hbc_reader.file_buffer.read(4), 'little')
                    + original_pos
                )

        yield result
